#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.core.phase_code import PhaseCode
from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.model.cim.iec61970.base.domain.unit_symbol import UnitSymbol

__all__ = ["Measurement", "Accumulator", "Analog", "Discrete"]


class Measurement(IdentifiedObject):
    """
    A Measurement represents any measured, calculated or non-measured non-calculated quantity. Any piece of equipment
    may contain Measurements, e.g. a substation may have temperature measurements and door open indications,
    a transformer may have oil temperature and tank pressure measurements, a bay may contain a number of power flow
    measurements and a Breaker may contain a switch status measurement.

    The PSR - Measurement association is intended to capture this use of Measurement and is included in the naming
    hierarchy based on EquipmentContainer. The naming hierarchy typically has Measurements as leafs,
    e.g. Substation-VoltageLevel-Bay-Switch-Measurement.

    Some Measurements represent quantities related to a particular sensor location in the network, e.g. a voltage
    transformer (PT) at a busbar or a current transformer (CT) at the bar between a breaker and an isolator. The
    sensing position is not captured in the PSR - Measurement association. Instead it is captured by the Measurement
    - Terminal association that is used to define the sensing location in the network topology. The location is
    defined by the connection of the Terminal to ConductingEquipment.

    If both a Terminal and PSR are associated, and the PSR is of type ConductingEquipment, the associated Terminal
    should belong to that ConductingEquipment instance.

    When the sensor location is needed both Measurement-PSR and Measurement-Terminal are used. The Measurement-Terminal
    association is never used alone.
    """

    power_system_resource_mrid: Optional[str] = None
    """The MRID of the power system resource that contains the measurement."""

    remote_source: Optional[RemoteSource] = None
    """The `zepben.evolve.cim.iec61970.base.scada.remote_source.RemoteSource` taking the `Measurement`"""

    terminal_mrid: Optional[str] = None
    """A measurement may be associated with a terminal in the network."""

    phases: PhaseCode = PhaseCode.ABC
    """Indicates to which phases the measurement applies and avoids the need to use 'measurementType' to also encode phase information 
    (which would explode the types). The phase information in Measurement, along with 'measurementType' and 'phases' uniquely defines a Measurement for a 
    device, based on normal network phase. Their meaning will not change when the computed energizing phasing is changed due to jumpers or other reasons. 
    If the attribute is missing three phases (ABC) shall be assumed."""

    unitSymbol: UnitSymbol = UnitSymbol.NONE
    """Specifies the type of measurement.  For example, this specifies if the measurement represents an indoor temperature, outdoor temperature, bus voltage, 
    line flow, etc. When the measurementType is set to "Specialization", the type of Measurement is defined in more detail by the specialized class which 
    inherits from Measurement."""


class Accumulator(Measurement):
    """Accumulator represents an accumulated (counted) Measurement, e.g. an energy value."""
    pass


class Analog(Measurement):
    """Analog represents an analog Measurement."""

    positive_flow_in: bool = False
    """If true then this measurement is an active power, reactive power or current with the convention that a positive value measured at the 
    Terminal means power is flowing into the related PowerSystemResource."""


class Discrete(Measurement):
    """Discrete represents a discrete Measurement, i.e. a Measurement representing discrete values, e.g. a Breaker position."""
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve import NetworkService, PerLengthSequenceImpedance, AcLineSegment, Pole, Streetlight
from zepben.evolve import resolver


def test_resolves_acls_plsi():
    ns = NetworkService()
    plsi = PerLengthSequenceImpedance()
    acls = AcLineSegment(per_length_sequence_impedance=plsi)
    br = resolver.per_length_sequence_impedance(acls)
    ns.resolve_or_defer_reference(br, plsi.mrid)
    assert plsi.mrid in ns.get_unresolved_reference_mrids(br)
    ns.add(acls)
    acls_fetched = ns.get(acls.mrid)
    assert acls_fetched.per_length_sequence_impedance == plsi

    ns = NetworkService()
    ns.add(plsi)
    ns.resolve_or_defer_reference(resolver.per_length_sequence_impedance(acls), plsi.mrid)
    assert len(list(ns.get_unresolved_reference_mrids(br))) == 0
    ns.add(acls)
    acls_fetched = ns.get(acls.mrid)
    assert acls_fetched.per_length_sequence_impedance == plsi


def test_resolves_pole_streetlight():
    ns = NetworkService()
    pole = Pole()
    streetlight = Streetlight(pole=pole)
    pole.add_streetlight(streetlight)
    br = resolver.streetlights(pole)
    ns.resolve_or_defer_reference(br, streetlight.mrid)
    assert streetlight.mrid in ns.get_unresolved_reference_mrids(br)
    ns.add(streetlight)
    assert len(list(ns.get_unresolved_reference_mrids(br))) == 0
    streetlight_fetched = ns.get(streetlight.mrid)
    assert streetlight == streetlight_fetched



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61970.base.core.power_system_resource import PowerSystemResource
from zepben.evolve.model.cim.iec61970.base.wires.energy_connection import EnergyConnection
from zepben.evolve.model.cim.iec61970.base.wires.phase_shunt_connection_kind import PhaseShuntConnectionKind
from zepben.evolve.model.cim.iec61970.base.wires.single_phase_kind import SinglePhaseKind

__all__ = ["EnergyConsumer", "EnergyConsumerPhase"]

from zepben.evolve.util import nlen, get_by_mrid, ngen, safe_remove


class EnergyConsumerPhase(PowerSystemResource):
    """A single phase of an energy consumer."""

    _energy_consumer: Optional[EnergyConsumer] = None

    phase: SinglePhaseKind = SinglePhaseKind.X
    """Phase of this energy consumer component. If the energy consumer is wye connected, the connection is from the indicated phase to the central ground or 
    neutral point. If the energy consumer is delta connected, the phase indicates an energy consumer connected from the indicated phase to the next
    logical non-neutral phase. """

    p: float = 0.0
    """Active power of the load. Load sign convention is used, i.e. positive sign means flow out from a node. For voltage dependent loads the value is at 
    rated voltage. Starting value for a steady state solution."""

    q: float = 0.0
    """Reactive power of the load. Load sign convention is used, i.e. positive sign means flow out from a node. For voltage dependent loads the value is at 
    rated voltage. Starting value for a steady state solution."""

    p_fixed: float = 0.0
    """Active power of the load that is a fixed quantity. Load sign convention is used, i.e. positive sign means flow out from a node."""

    q_fixed: float = 0.0
    """Reactive power of the load that is a fixed quantity. Load sign convention is used, i.e. positive sign means flow out from a node."""

    def __init__(self, energy_consumer: EnergyConsumer = None):
        if energy_consumer:
            self.energy_consumer = energy_consumer

    @property
    def energy_consumer(self):
        """The `zepben.evolve.cim.iec61970.base.wires.EnergyConsumer` that has this phase."""
        return self._energy_consumer

    @energy_consumer.setter
    def energy_consumer(self, ec):
        if self._energy_consumer is None or self._energy_consumer is ec:
            self._energy_consumer = ec
        else:
            raise ValueError(f"energy_consumer for {str(self)} has already been set to {self._energy_consumer}, cannot reset this field to {ec}")


class EnergyConsumer(EnergyConnection):
    """Generic user of energy - a point of consumption on the power system phases. May also represent a pro-sumer with negative p/q values. """

    _energy_consumer_phases: Optional[List[EnergyConsumerPhase]] = None
    """The individual phase models for this energy consumer."""

    customer_count: int = 0
    """Number of individual customers represented by this demand."""

    grounded: bool = False
    """Used for Yn and Zn connections. True if the neutral is solidly grounded."""

    phase_connection: PhaseShuntConnectionKind = PhaseShuntConnectionKind.D
    """`zepben.protobuf.cim.iec61970.base.wires.phase_shunt_connection_kind.PhaseShuntConnectionKind` - The type of phase connection, 
    such as wye, delta, I (single phase)."""

    p: float = 0.0
    """Active power of the load. Load sign convention is used, i.e. positive sign means flow out from a node. For voltage dependent loads the value is at 
    rated voltage. Starting value for a steady state solution."""

    p_fixed: float = 0.0
    """Active power of the load that is a fixed quantity. Load sign convention is used, i.e. positive sign means flow out from a node."""

    q: float = 0.0
    """Reactive power of the load. Load sign convention is used, i.e. positive sign means flow out from a node. For voltage dependent loads the value is at 
    rated voltage. Starting value for a steady state solution."""

    q_fixed: float = 0.0
    """Power of the load that is a fixed quantity. Load sign convention is used, i.e. positive sign means flow out from a node."""

    def __init__(self, usage_points: List[UsagePoint] = None, equipment_containers: List[EquipmentContainer] = None,
                 operational_restrictions: List[OperationalRestriction] = None, current_feeders: List[Feeder] = None, terminals: List[Terminal] = None,
                 energy_consumer_phases: List[EnergyConsumerPhase] = None):
        super(EnergyConsumer, self).__init__(usage_points=usage_points, equipment_containers=equipment_containers, operational_restrictions=operational_restrictions,
                         current_feeders=current_feeders, terminals=terminals)
        if energy_consumer_phases:
            for phase in energy_consumer_phases:
                self.add_phase(phase)

    def has_phases(self):
        """
        Check if this consumer has any associated `EnergyConsumerPhases`
        Returns True if there is at least one `EnergyConsumerPhase`, otherwise False
        """
        return nlen(self._energy_consumer_phases) > 0

    def num_phases(self):
        """Get the number of `EnergySourcePhase`s for this `EnergyConsumer`."""
        return nlen(self._energy_consumer_phases)

    @property
    def phases(self) -> Generator[EnergyConsumerPhase, None, None]:
        """The individual phase models for this energy consumer."""
        return ngen(self._energy_consumer_phases)

    def get_phase(self, mrid: str) -> EnergyConsumer:
        """
        Get the `EnergyConsumerPhase` for this `EnergyConsumer` identified by `mrid`

        `mrid` The mRID of the required `EnergyConsumerPhase`
        Returns The `EnergyConsumerPhase` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._energy_consumer_phases, mrid)

    def add_phase(self, phase: EnergyConsumerPhase) -> EnergyConsumer:
        """
        Associate an `EnergyConsumerPhase` with this `EnergyConsumer`

        `phase` the `EnergyConsumerPhase` to associate with this `EnergyConsumer`.
        Returns A reference to this `EnergyConsumer` to allow fluent use.
        Raises `ValueError` if another `EnergyConsumerPhase` with the same `mrid` already exists for this `EnergyConsumer`.
        """
        if self._validate_reference(phase, self.get_phase, "An EnergyConsumerPhase"):
            return self
        self._energy_consumer_phases = list() if self._energy_consumer_phases is None else self._energy_consumer_phases
        self._energy_consumer_phases.append(phase)
        return self

    def remove_phase(self, phase: EnergyConsumerPhase) -> EnergyConsumer:
        """
        Disassociate `phase` from this `OperationalRestriction`.

        `phase` the `EnergyConsumerPhase` to disassociate with this `EnergyConsumer`.
        Raises `KeyError` if `phase` was not associated with this `EnergyConsumer`.
        Returns A reference to this `EnergyConsumer` to allow fluent use.
        Raises `ValueError` if `phase` was not associated with this `EnergyConsumer`.
        """
        self._energy_consumer_phases = safe_remove(self._energy_consumer_phases, phase)
        return self

    def clear_phases(self) -> EnergyConsumer:
        """
        Clear all phases.
        Returns A reference to this `EnergyConsumer` to allow fluent use.
        """
        self._energy_consumer_phases = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from dataclassy import dataclass
from operator import attrgetter

from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment
from zepben.evolve.model.cim.iec61970.base.core.terminal import Terminal
from zepben.evolve.model.phases import NominalPhasePath
from typing import List, Optional, Tuple

__all__ = ["ConnectivityResult", "get_connectivity", "terminal_compare", "get_connected_equipment"]


def terminal_compare(terminal: Terminal, other: Terminal):
    """
    This definition should only be used for sorting within a `zepben.evolve.traversals.queue.PriorityQueue`
    `terminal` The terminal to compare
    `other` The terminal to compare against
    Returns True if `terminal` has more phases than `other`, False otherwise.
    """
    return terminal.phases.num_phases > other.phases.num_phases


Terminal.__lt__ = terminal_compare


def get_connectivity(terminal: Terminal, phases: Set[SinglePhaseKind] = None, exclude=None):
    """
    Get the connectivity between this terminal and all other terminals in its `ConnectivityNode`.
    `cores` Core paths to trace between the terminals. Defaults to all cores.
    `exclude` `zepben.evolve.iec61970.base.core.terminal.Terminal`'s to exclude from the result. Will be skipped if encountered.
    Returns List of `ConnectivityResult`'s for this terminal.
    """
    if exclude is None:
        exclude = set()
    if phases is None:
        phases = terminal.phases.single_phases
    trace_phases = phases.intersection(terminal.phases.single_phases)
    cn = terminal.connectivity_node if terminal.connectivity_node else []
    results = []
    for term in cn:
        if terminal is not term and term not in exclude:  # Don't include ourselves, or those specifically excluded.
            cr = _terminal_connectivity(terminal, term, trace_phases)
            if cr.nominal_phase_paths:
                results.append(cr)
    return results


Terminal.connected_terminals = get_connectivity


def get_connected_equipment(cond_equip, exclude: Set = None):
    """
    Get all `ConductingEquipment` connected to this piece of equipment. An `Equipment` is connected if it has
    a `zepben.evolve.iec61970.base.core.terminal.Terminal` associated with a `ConnectivityNode` that this `ConductingEquipment` is also associated with.

    `exclude` Equipment to exclude from return.
    Returns A list of `ConductingEquipment` that are connected to this.
    """
    if exclude is None:
        exclude = []
    connected_equip = []
    for terminal in cond_equip._terminals:
        conn_node = terminal.connectivity_node
        for term in conn_node:
            if term.conducting_equipment in exclude:
                continue
            if term != terminal:  # Don't include ourselves.
                connected_equip.append(term.conducting_equipment)
    return connected_equip


ConductingEquipment.connected_equipment = get_connected_equipment


def _terminal_connectivity(terminal: Terminal, connected_terminal: Terminal, phases: Set[SinglePhaseKind]) -> ConnectivityResult:
    nominal_phase_paths = [NominalPhasePath(phase, phase) for phase in phases if phase in connected_terminal.phases.single_phases]

    if not nominal_phase_paths:
        xy_phases = {phase for phase in phases if phase == SinglePhaseKind.X or phase == SinglePhaseKind.Y}
        connected_xy_phases = {phase for phase in connected_terminal.phases.single_phases if phase == SinglePhaseKind.X or phase == SinglePhaseKind.Y}

        _process_xy_phases(terminal, connected_terminal, phases, xy_phases, connected_xy_phases, nominal_phase_paths)

    return ConnectivityResult(from_terminal=terminal, to_terminal=connected_terminal, nominal_phase_paths=nominal_phase_paths)


def _process_xy_phases(terminal: Terminal, connected_terminal: Terminal, phases: Set[SinglePhaseKind], xy_phases: Set[SinglePhaseKind],
                       connectied_xy_phases: Set[SinglePhaseKind], nominal_phase_paths: List[NominalPhasePath]):
    if (not xy_phases and not connectied_xy_phases) or (xy_phases and connectied_xy_phases):
        return
    for phase in xy_phases:
        i = terminal.phases.single_phases.index(phase)
        if i < len(connected_terminal.phases.single_phases):
            nominal_phase_paths.append(NominalPhasePath(from_phase=phase, to_phase=connected_terminal.phases.single_phases[i]))

    for phase in connectied_xy_phases:
        i = connected_terminal.phases.single_phases.index(phase)
        if i < len(terminal.phases.single_phases):
            terminal_phase = terminal.phases.single_phases[i]
            if terminal_phase in phases:
                nominal_phase_paths.append(NominalPhasePath(from_phase=terminal_phase, to_phase=phase))


@dataclass(slots=True)
class ConnectivityResult(object):
    """
    Stores the connectivity between two terminals, including the mapping between the nominal phases.
    This class is intended to be used in an immutable way. You should avoid modifying it after it has been created.
    """

    from_terminal: Terminal
    """The terminal from which the connectivity was requested."""

    to_terminal: Terminal
    """The terminal which is connected to the requested terminal."""

    nominal_phase_paths: Tuple[NominalPhasePath]
    """The mapping of nominal phase paths between the from and to terminals."""

    def __init__(self, nominal_phase_paths: List[NominalPhasePath]):
        self.nominal_phase_paths = tuple(sorted(nominal_phase_paths, key=attrgetter('from_terminal', 'to_terminal')))

    def __eq__(self, other: ConnectivityResult):
        if self is other:
            return True
        try:
            return self.from_terminal is other.from_terminal and self.to_terminal is other.to_terminal and self.nominal_phase_paths != other.nominal_phase_paths
        except:
            return False

    def __ne__(self, other):
        if self is other:
            return False
        try:
            return self.from_terminal is not other.from_terminal or self.to_terminal is not other.to_terminal or self.nominal_phase_paths != other.nominal_phase_paths
        except:
            return True

    def __str__(self):
        return (f"ConnectivityResult(from_terminal={self.from_equip.mrid}-t{self.from_terminal.sequence_number}"
                f", to_terminal={self.to_equip.mrid}-t{self.to_terminal.sequence_number}, core_paths={self.nominal_phase_paths})")

    def __hash__(self):
        res = self.from_terminal.mrid.__hash__()
        res = 31 * res + self.to_terminal.mrid.__hash__()
        res = 31 * res + self.nominal_phase_paths.__hash__()
        return res

    @property
    def from_equip(self) -> Optional[ConductingEquipment]:
        """The conducting equipment that owns the `from_terminal."""
        return self.from_terminal.conducting_equipment

    @property
    def to_equip(self) -> Optional[ConductingEquipment]:
        """The conducting equipment that owns the `to_terminal`."""
        return self.to_terminal.conducting_equipment

    @property
    def from_nominal_phases(self) -> List[SinglePhaseKind]:
        """The nominal phases that are connected in the `from_terminal`."""
        return [npp.from_phase for npp in self.nominal_phase_paths]

    @property
    def to_nominal_phases(self) -> List[SinglePhaseKind]:
        """The nominal phases that are connected in the `to_terminal`."""
        return [npp.to_phase for npp in self.nominal_phase_paths]





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from dataclassy import dataclass
from typing import Callable, Optional


from zepben.evolve.model import Organisation
from zepben.evolve.model import OrganisationRole
from zepben.evolve.model import WireInfo
from zepben.evolve.model import Asset
from zepben.evolve.model import Pole
from zepben.evolve.model import Streetlight
from zepben.evolve.model import AssetOrganisationRole
from zepben.evolve.model import Location
from zepben.evolve.model import Customer
from zepben.evolve.model import CustomerAgreement
from zepben.evolve.model import PricingStructure
from zepben.evolve.model import Tariff
from zepben.evolve.model import EndDevice, UsagePoint
from zepben.evolve.model import OperationalRestriction
from zepben.evolve.model import AuxiliaryEquipment
from zepben.evolve.model import BaseVoltage
from zepben.evolve.model import ConductingEquipment
from zepben.evolve.model import ConnectivityNode
from zepben.evolve.model import IdentifiedObject
from zepben.evolve.model import Equipment
from zepben.evolve.model import Terminal
from zepben.evolve.model import Diagram, DiagramObject
from zepben.evolve.model import Measurement
from zepben.evolve.model import Control
from zepben.evolve.model import RemoteSource
from zepben.evolve.model import RemoteControl
from zepben.evolve.model import AcLineSegment, Conductor
from zepben.evolve.model import EnergyConsumer, EnergyConsumerPhase
from zepben.evolve.model import EnergySource
from zepben.evolve.model import EnergySourcePhase
from zepben.evolve.model import PerLengthSequenceImpedance, Feeder, EquipmentContainer, Substation, GeographicalRegion, SubGeographicalRegion, \
    PowerSystemResource, PowerTransformer, PowerTransformerEnd, RatioTapChanger, TransformerEnd

__all__ = ["acls_to_plsi_resolver", "asset_to_asset_org_role_resolver", "asset_to_location_resolver", "pole_to_streetlight_resolver",
           "streetlight_to_pole_resolver", "aux_equip_to_term_resolver", "cond_equip_to_bv_resolver", "cond_equip_to_terminal_resolver",
           "conductor_to_wire_info_resolver", "conn_node_to_term_resolver", "control_to_remote_control_resolver", "cust_to_custagr_resolver",
           "custagr_to_cust_resolver", "custagr_to_ps_resolver", "diag_to_diagobj_resolver", "diagobj_to_diag_resolver", "ed_to_up_resolver",
           "ed_to_loc_resolver", "ec_to_ecp_resolver", "ecp_to_ec_resolver", "es_to_esp_resolver", "esp_to_es_resolver", "eq_to_curfeeder_resolver",
           "eq_to_ec_resolver", "eq_to_or_resolver", "eq_to_up_resolver", "ec_to_eq_resolver", "curfeeder_to_eq_resolver", "feeder_to_nes_resolver",
           "feeder_to_nht_resolver", "gr_to_sgr_resolver", "meas_to_rs_resolver", "or_to_eq_resolver", "orgr_to_org_resolver", "psr_to_loc_resolver",
           "pt_to_pte_resolver", "pte_to_pt_resolver", "ps_to_tariff_resolver", "rtc_to_te_resolver", "rc_to_cont_resolver", "rs_to_meas_resolver",
           "sgr_to_gr_resolver", "sgr_to_sub_resolver", "sub_to_feeder_resolver", "sub_to_sgr_resolver", "sub_to_circuit_resolver", "sub_to_eloop_resolver",
           "sub_to_loop_resolver", "term_to_ce_resolver", "term_to_cn_resolver", "te_to_term_resolver", "te_to_bv_resolver", "te_to_rtc_resolver",
           "up_to_ed_resolver", "up_to_eq_resolver", "up_to_loc_resolver", "circuit_to_loop_resolver", "circuit_to_sub_resolver", "circuit_to_term_resolver",
           "loop_to_circuit_resolver", "loop_to_esub_resolver", "loop_to_sub_resolver", "BoundReferenceResolver", "ReferenceResolver", "UnresolvedReference"]

from zepben.evolve.model import Circuit, Loop


@dataclass(frozen=True, eq=False, slots=True)
class ReferenceResolver(object):
    from_class: type
    to_class: type
    resolve: Callable[[IdentifiedObject, IdentifiedObject], None]

    def __eq__(self, other):
        return self.from_class is other.from_class and self.to_class is other.to_class and self.resolve is other.resolve

    def __neq__(self, other):
        return self.from_class is not other.from_class or self.to_class is not other.to_class or self.resolve is not other.resolve


@dataclass(frozen=True, eq=False, slots=True)
class BoundReferenceResolver(object):
    from_obj: IdentifiedObject
    resolver: ReferenceResolver
    reverse_resolver: Optional[ReferenceResolver]

    def __eq__(self, other):
        return self.from_obj is other.from_obj and self.resolver is other.resolver

    def __neq__(self, other):
        return self.from_obj is not other.from_obj or self.resolver is not other.resolver


@dataclass(frozen=True, eq=False, slots=True)
class UnresolvedReference(object):
    from_ref: IdentifiedObject
    to_mrid: str
    resolver: ReferenceResolver


def _resolve_ce_term(ce, t):
    t.conducting_equipment = ce
    ce.add_terminal(t)


def _resolve_pt_pte(pt, pte):
    pte.power_transformer = pt
    pt.add_end(pte)


def _resolve_diag_diagobj(diag, diag_obj):
    diag_obj.diagram = diag
    diag.add_diagram_object(diag_obj)


acls_to_plsi_resolver = ReferenceResolver(AcLineSegment, PerLengthSequenceImpedance, lambda t, r: setattr(t, 'per_length_sequence_impedance', r))

asset_to_asset_org_role_resolver = ReferenceResolver(Asset, AssetOrganisationRole, lambda t, r: t.add_organisation_role(r))
asset_to_location_resolver = ReferenceResolver(Asset, Location, lambda t, r: setattr(t, 'location', r))

pole_to_streetlight_resolver = ReferenceResolver(Pole, Streetlight, lambda t, r: t.add_streetlight(r))
streetlight_to_pole_resolver = ReferenceResolver(Streetlight, Pole, lambda t, r: setattr(t, 'pole', r))

aux_equip_to_term_resolver = ReferenceResolver(AuxiliaryEquipment, Terminal, lambda t, r: setattr(t, 'terminal', r))

cond_equip_to_bv_resolver = ReferenceResolver(ConductingEquipment, BaseVoltage, lambda t, r: setattr(t, 'base_voltage', r))
cond_equip_to_terminal_resolver = ReferenceResolver(ConductingEquipment, Terminal, _resolve_ce_term)

conductor_to_wire_info_resolver = ReferenceResolver(Conductor, WireInfo, lambda t, r: setattr(t, 'asset_info', r))

conn_node_to_term_resolver = ReferenceResolver(ConnectivityNode, Terminal, lambda t, r: t.add_terminal(r))

control_to_remote_control_resolver = ReferenceResolver(Control, RemoteControl, lambda t, r: setattr(t, 'remote_control', r))

cust_to_custagr_resolver = ReferenceResolver(Customer, CustomerAgreement, lambda t, r: t.add_agreement(r))
custagr_to_cust_resolver = ReferenceResolver(CustomerAgreement, Customer, lambda t, r: setattr(t, 'customer', r))

custagr_to_ps_resolver = ReferenceResolver(CustomerAgreement, PricingStructure, lambda t, r: t.add_pricing_structure(r))

diag_to_diagobj_resolver = ReferenceResolver(Diagram, DiagramObject, _resolve_diag_diagobj)
diagobj_to_diag_resolver = ReferenceResolver(DiagramObject, Diagram, lambda t, r: setattr(t, 'diagram', r))

ed_to_up_resolver = ReferenceResolver(EndDevice, UsagePoint, lambda t, r: t.add_usage_point(r))
ed_to_loc_resolver = ReferenceResolver(EndDevice, Location, lambda t, r: setattr(t, 'service_location', r))

ec_to_ecp_resolver = ReferenceResolver(EnergyConsumer, EnergyConsumerPhase, lambda t, r: t.add_phase(r))
ecp_to_ec_resolver = ReferenceResolver(EnergyConsumerPhase, EnergyConsumer, lambda t, r: setattr(t, 'energy_consumer', r))

es_to_esp_resolver = ReferenceResolver(EnergySource, EnergySourcePhase, lambda t, r: t.add_phase(r))
esp_to_es_resolver = ReferenceResolver(EnergySourcePhase, EnergySource, lambda t, r: setattr(t, 'energy_source', r))

eq_to_curfeeder_resolver = ReferenceResolver(Equipment, Feeder, lambda t, r: t.add_current_feeder(r))
eq_to_ec_resolver = ReferenceResolver(Equipment, EquipmentContainer, lambda t, r: t.add_container(r))
eq_to_or_resolver = ReferenceResolver(Equipment, OperationalRestriction, lambda t, r: t.add_restriction(r))
eq_to_up_resolver = ReferenceResolver(Equipment, UsagePoint, lambda t, r: t.add_usage_point(r))

ec_to_eq_resolver = ReferenceResolver(EquipmentContainer, Equipment, lambda t, r: t.add_equipment(r))

curfeeder_to_eq_resolver = ReferenceResolver(Feeder, Equipment, lambda t, r: t.add_equipment(r))

feeder_to_nes_resolver = ReferenceResolver(Feeder, Substation, lambda t, r: setattr(t, 'normal_energizing_substation', r))
feeder_to_nht_resolver = ReferenceResolver(Feeder, Terminal, lambda t, r: setattr(t, 'normal_head_terminal', r))

gr_to_sgr_resolver = ReferenceResolver(GeographicalRegion, SubGeographicalRegion, lambda t, r: t.add_sub_geographical_region(r))

meas_to_rs_resolver = ReferenceResolver(Measurement, RemoteSource, lambda t, r: setattr(t, 'remote_source', r))

or_to_eq_resolver = ReferenceResolver(OperationalRestriction, Equipment, lambda t, r: t.add_equipment(r))

orgr_to_org_resolver = ReferenceResolver(OrganisationRole, Organisation, lambda t, r: setattr(t, 'organisation', r))

psr_to_loc_resolver = ReferenceResolver(PowerSystemResource, Location, lambda t, r: setattr(t, 'location', r))

pt_to_pte_resolver = ReferenceResolver(PowerTransformer, PowerTransformerEnd, _resolve_pt_pte)
pte_to_pt_resolver = ReferenceResolver(PowerTransformerEnd, PowerTransformer, lambda t, r: setattr(t, 'power_transformer', r))

ps_to_tariff_resolver = ReferenceResolver(PricingStructure, Tariff, lambda t, r: t.tariff_to_cim(r))

rtc_to_te_resolver = ReferenceResolver(RatioTapChanger, PowerTransformerEnd, lambda t, r: setattr(t, 'transformer_end', r))

rc_to_cont_resolver = ReferenceResolver(RemoteControl, Control, lambda t, r: setattr(t, 'control', r))

rs_to_meas_resolver = ReferenceResolver(RemoteSource, Measurement, lambda t, r: setattr(t, 'measurement', r))

sgr_to_gr_resolver = ReferenceResolver(SubGeographicalRegion, GeographicalRegion, lambda t, r: setattr(t, 'geographical_region', r))
sgr_to_sub_resolver = ReferenceResolver(SubGeographicalRegion, Substation, lambda t, r: t.add_substation(r))

sub_to_feeder_resolver = ReferenceResolver(Substation, Feeder, lambda t, r: t.add_feeder(r))
sub_to_sgr_resolver = ReferenceResolver(Substation, SubGeographicalRegion, lambda t, r: setattr(t, 'sub_geographical_region', r))
sub_to_circuit_resolver = ReferenceResolver(Substation, Circuit, lambda t, r: t.add_circuit(r))
sub_to_loop_resolver = ReferenceResolver(Substation, Loop, lambda t, r: t.add_loop(r))
sub_to_eloop_resolver = ReferenceResolver(Substation, Loop, lambda t, r: t.add_energized_loop(r))

term_to_ce_resolver = ReferenceResolver(Terminal, ConductingEquipment, lambda t, r: setattr(t, 'conducting_equipment', r))
term_to_cn_resolver = ReferenceResolver(Terminal, ConnectivityNode, lambda t, r: setattr(t, 'connectivity_node', r))

te_to_term_resolver = ReferenceResolver(TransformerEnd, Terminal, lambda t, r: setattr(t, 'terminal', r))
te_to_bv_resolver = ReferenceResolver(TransformerEnd, BaseVoltage, lambda t, r: setattr(t, 'base_voltage', r))
te_to_rtc_resolver = ReferenceResolver(TransformerEnd, RatioTapChanger, lambda t, r: setattr(t, 'ratio_tap_changer', r))

up_to_ed_resolver = ReferenceResolver(UsagePoint, EndDevice, lambda t, r: t.add_end_device(r))
up_to_eq_resolver = ReferenceResolver(UsagePoint, Equipment, lambda t, r: t.add_equipment(r))
up_to_loc_resolver = ReferenceResolver(UsagePoint, Location, lambda t, r: setattr(t, 'usage_point_location', r))

circuit_to_term_resolver = ReferenceResolver(Circuit, Terminal, lambda t, r: t.add_terminal(r))
circuit_to_loop_resolver = ReferenceResolver(Circuit, Loop, lambda t, r: setattr(t, 'loop', r))
circuit_to_sub_resolver = ReferenceResolver(Circuit, Substation, lambda t, r: t.add_substation(r))

loop_to_circuit_resolver = ReferenceResolver(Loop, Circuit, lambda t, r: t.add_circuit(r))
loop_to_sub_resolver = ReferenceResolver(Loop, Substation, lambda t, r: t.add_substation(r))
loop_to_esub_resolver = ReferenceResolver(Loop, Substation, lambda t, r: t.add_energizing_substation(r))





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import List, Optional, Generator

from zepben.evolve.model.cim.iec61970.base.core.conducting_equipment import ConductingEquipment
from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.model.cim.iec61970.base.core.power_system_resource import PowerSystemResource
from zepben.evolve.model.cim.iec61970.base.wires.vector_group import VectorGroup
from zepben.evolve.model.cim.iec61970.base.wires.winding_connection import WindingConnection
from zepben.evolve.util import require, nlen, get_by_mrid, ngen, safe_remove

__all__ = ["TapChanger", "RatioTapChanger", "PowerTransformer", "PowerTransformerEnd", "TransformerEnd"]


class TapChanger(PowerSystemResource):
    """
    Mechanism for changing transformer winding tap positions.
    """

    control_enabled: bool = True
    """Specifies the regulation status of the equipment.  True is regulating, false is not regulating."""

    neutral_u: int = 0
    """Voltage at which the winding operates at the neutral tap setting."""

    _high_step: int = 1
    _low_step: int = 0
    _neutral_step: int = 0
    _normal_step: int = 0
    _step: float = 0.0

    def __init__(self, high_step: int = 1, low_step: int = 0, neutral_step: int = 0, normal_step: int = 0, step: float = 0.0):
        self._high_step = high_step
        self._low_step = low_step
        self._neutral_step = neutral_step
        self._normal_step = normal_step
        self._step = step
        self._validate_steps()

    @property
    def high_step(self):
        """Highest possible tap step position, advance from neutral. The attribute shall be greater than lowStep."""
        return self._high_step

    @high_step.setter
    def high_step(self, val):
        require(val > self._low_step, lambda: f"High step {val} must be greater than low step {self._low_step}")
        self._check_steps(self.low_step, val)
        self._high_step = val

    @property
    def low_step(self):
        """Lowest possible tap step position, retard from neutral"""
        return self._low_step

    @low_step.setter
    def low_step(self, val):
        require(val < self._high_step, lambda: f"Low step {val} must be less than high step {self._high_step}")
        self._check_steps(val, self.high_step)
        self._low_step = val

    @property
    def neutral_step(self):
        """The neutral tap step position for this winding. The attribute shall be equal or greater than lowStep and equal or less than highStep."""
        return self._neutral_step

    @neutral_step.setter
    def neutral_step(self, val):
        require(self._low_step <= val <= self._high_step, lambda: f"Neutral step {val} must be between high step {self._high_step} and low step {self._low_step}")
        self._neutral_step = val

    @property
    def normal_step(self):
        """
        The tap step position used in "normal" network operation for this winding. For a "Fixed" tap changer indicates the current physical tap setting.
        The attribute shall be equal or greater than lowStep and equal or less than highStep.
        """
        return self._normal_step

    @normal_step.setter
    def normal_step(self, val):
        require(self._low_step <= val <= self._high_step, lambda: f"Normal step {val} must be between high step {self._high_step} and low step {self._low_step}")
        self._normal_step = val

    @property
    def step(self):
        """
        Tap changer position. Starting step for a steady state solution. Non integer values are allowed to support continuous tap variables.
        The reasons for continuous value are to support study cases where no discrete tap changers has yet been designed, a solutions where a narrow voltage
        band force the tap step to oscillate or accommodate for a continuous solution as input.
        The attribute shall be equal or greater than lowStep and equal or less than highStep.
        """
        return self._step

    @step.setter
    def step(self, val):
        require(self._low_step <= val <= self._high_step, lambda: f"Step {val} must be between high step {self._high_step} and low step {self._low_step}")
        self._step = val

    def _check_steps(self, low, high):
        require(low <= self.step <= high, lambda: f"New value would invalidate current step of {self.step}")
        require(low <= self.normal_step <= high,
                lambda: f"New value would invalidate current normal_step of {self.normal_step}")
        require(low <= self.neutral_step <= high,
                lambda: f"New value would invalidate current neutral_step of {self.neutral_step}")

    def _validate_steps(self):
        require(self.high_step > self.low_step,
                lambda: f"High step [{self.high_step}] must be greater than low step [{self.low_step}]")
        require(self.low_step <= self.neutral_step <= self.high_step,
                lambda: f"Neutral step [{self.neutral_step}] must be between high step [{self._high_step}] and low step [{self._low_step}]")
        require(self._low_step <= self.normal_step <= self._high_step,
                lambda: f"Normal step [{self.normal_step}] must be between high step [{self._high_step}] and low step [{self._low_step}]")
        require(self._low_step <= self.step <= self._high_step,
                lambda: f"Step [{self.step}] must be between high step [{self._high_step}] and low step [{self._low_step}]")


class RatioTapChanger(TapChanger):
    """
    A tap changer that changes the voltage ratio impacting the voltage magnitude but not the phase angle across the transformer.

    Angle sign convention (general): Positive value indicates a positive phase shift from the winding where the tap is located to the other winding
    (for a two-winding transformer).
    """

    transformer_end: Optional[TransformerEnd] = None
    """`TransformerEnd` to which this ratio tap changer belongs."""

    step_voltage_increment: float = 0.0
    """Tap step increment, in per cent of neutral voltage, per step position."""


class TransformerEnd(IdentifiedObject):
    """
    A conducting connection point of a power transformer. It corresponds to a physical transformer winding terminal.
    In earlier CIM versions, the TransformerWinding class served a similar purpose, but this class is more flexible
    because it associates to terminal but is not a specialization of ConductingEquipment.
    """
    grounded: bool = False
    """(for Yn and Zn connections) True if the neutral is solidly grounded."""

    r_ground: float = 0.0
    """(for Yn and Zn connections) Resistance part of neutral impedance where 'grounded' is true"""

    x_ground: float = 0.0
    """(for Yn and Zn connections) Reactive part of neutral impedance where 'grounded' is true"""

    ratio_tap_changer: Optional[RatioTapChanger] = None
    """Ratio tap changer associated with this transformer end."""

    terminal: Optional[Terminal] = None
    """The terminal of the transformer that this end is associated with"""

    base_voltage: Optional[BaseVoltage] = None
    """Base voltage of the transformer end.  This is essential for PU calculation."""

    end_number: int = 0
    """Number for this transformer end, corresponding to the end’s order in the power transformer vector group or phase angle clock number. 
    Highest voltage winding should be 1. Each end within a power transformer should have a unique subsequent end number. 
    Note the transformer end number need not match the terminal sequence number."""


class PowerTransformerEnd(TransformerEnd):
    """
    A PowerTransformerEnd is associated with each Terminal of a PowerTransformer.

    The impedance values r, r0, x, and x0 of a PowerTransformerEnd represents a star equivalent as follows

    1) for a two Terminal PowerTransformer the high voltage PowerTransformerEnd has non zero values on r, r0, x, and x0
    while the low voltage PowerTransformerEnd has zero values for r, r0, x, and x0.
    2) for a three Terminal PowerTransformer the three PowerTransformerEnds represents a star equivalent with each leg
    in the star represented by r, r0, x, and x0 values.
    3) For a three Terminal transformer each PowerTransformerEnd shall have g, g0, b and b0 values corresponding the no load losses
    distributed on the three PowerTransformerEnds. The total no load loss shunt impedances may also be placed at one of the
    PowerTransformerEnds, preferably the end numbered 1, having the shunt values on end 1 is the preferred way.
    4) for a PowerTransformer with more than three Terminals the PowerTransformerEnd impedance values cannot be used.
    Instead use the TransformerMeshImpedance or split the transformer into multiple PowerTransformers.
    """

    _power_transformer: Optional[PowerTransformer] = None
    """The power transformer of this power transformer end."""

    rated_s: int = 0
    """Normal apparent power rating. The attribute shall be a positive value. For a two-winding transformer the values for the high and low voltage sides 
    shall be identical."""

    rated_u: int = 0
    """Rated voltage: phase-phase for three-phase windings, and either phase-phase or phase-neutral for single-phase windings. A high voltage side, as given by 
    TransformerEnd.endNumber, shall have a ratedU that is greater or equal than ratedU for the lower voltage sides."""

    r: float = 0.0
    """Resistance (star-phases) of the transformer end. The attribute shall be equal or greater than zero for non-equivalent transformers."""

    x: float = 0.0
    """Positive sequence series reactance (star-phases) of the transformer end."""

    r0: float = 0.0
    """Zero sequence series resistance (star-phases) of the transformer end."""

    x0: float = 0.0
    """Zero sequence series reactance of the transformer end."""

    g: float = 0.0
    """Magnetizing branch conductance."""

    g0: float = 0.0
    """Zero sequence magnetizing branch conductance (star-phases)."""

    b: float = 0.0
    """Magnetizing branch susceptance (B mag).  The value can be positive or negative."""

    b0: float = 0.0
    """Zero sequence magnetizing branch susceptance."""

    connection_kind: WindingConnection = WindingConnection.UNKNOWN_WINDING
    """Kind of `zepben.protobuf.cim.iec61970.base.wires.winding_connection.WindingConnection` for this end."""

    phase_angle_clock: int = 0
    """Terminal voltage phase angle displacement where 360 degrees are represented with clock hours. The valid values are 0 to 11. For example, for the 
    secondary side end of a transformer with vector group code of 'Dyn11', specify the connection kind as wye with neutral and specify the phase angle of the 
    clock as 11. The clock value of the transformer end number specified as 1, is assumed to be zero."""

    def __init__(self, power_transformer: PowerTransformer = None):
        if power_transformer:
            self.power_transformer = power_transformer

    @property
    def power_transformer(self):
        """The power transformer of this power transformer end."""
        return self._power_transformer

    @power_transformer.setter
    def power_transformer(self, pt):
        if self._power_transformer is None or self._power_transformer is pt:
            self._power_transformer = pt
        else:
            raise ValueError(f"power_transformer for {str(self)} has already been set to {self._power_transformer}, cannot reset this field to {pt}")

    @property
    def nominal_voltage(self):
        return self.base_voltage.nominal_voltage if self.base_voltage else self.rated_u


class PowerTransformer(ConductingEquipment):
    """
    An electrical device consisting of  two or more coupled windings, with or without a magnetic core, for introducing
    mutual coupling between electric circuits.

    Transformers can be used to control voltage and phase shift (active power flow). A power transformer may be composed of separate transformer tanks that
    need not be identical. A power transformer can be modeled with or without tanks and is intended for use in both balanced and unbalanced representations.

    A power transformer typically has two terminals, but may have one (grounding), three or more terminals.

    The inherited association ConductingEquipment.BaseVoltage should not be used.
    The association from TransformerEnd to BaseVoltage should be used instead.

    Attributes -
        vector_group : `zepben.protobuf.cim.iec61970.base.wires.VectorGroup` of the transformer for protective relaying.
        power_transformer_ends : 
                                 
                                 
    """
    vector_group: VectorGroup = VectorGroup.UNKNOWN
    """
    Vector group of the transformer for protective relaying, e.g., Dyn1. For unbalanced transformers, this may not be simply
    determined from the constituent winding connections and phase angle displacements.
                                                                                                                            
    The vectorGroup string consists of the following components in the order listed: high voltage winding connection, mid
    voltage winding connection(for three winding transformers), phase displacement clock number from 0 to 11,  low voltage
    winding connection phase displacement clock number from 0 to 11.   The winding connections are D(delta), Y(wye),
    YN(wye with neutral), Z(zigzag), ZN(zigzag with neutral), A(auto transformer). Upper case means the high voltage,
    lower case mid or low.The high voltage winding always has clock position 0 and is not included in the vector group
    string.  Some examples: YNy0(two winding wye to wye with no phase displacement), YNd11(two winding wye to delta with
    330 degrees phase displacement), YNyn0d5(three winding transformer wye with neutral high voltage, wye with neutral mid
    voltage and no phase displacement, delta low voltage with 150 degrees displacement).
                                                                                                                            
    Phase displacement is defined as the angular difference between the phasors representing the voltages between the
    neutral point(real or imaginary) and the corresponding terminals of two windings, a positive sequence voltage system
    being applied to the high-voltage terminals, following each other in alphabetical sequence if they are lettered, or in
    numerical sequence if they are numbered: the phasors are assumed to rotate in a counter-clockwise sense.
    """

    _power_transformer_ends: Optional[List[PowerTransformerEnd]] = None

    def __init__(self, usage_points: List[UsagePoint] = None, equipment_containers: List[EquipmentContainer] = None,
                 operational_restrictions: List[OperationalRestriction] = None, current_feeders: List[Feeder] = None, terminals: List[Terminal] = None,
                 power_transformer_ends: List[PowerTransformerEnd] = None):
        super(PowerTransformer, self).__init__(usage_points=usage_points, equipment_containers=equipment_containers, operational_restrictions=operational_restrictions,
                         current_feeders=current_feeders, terminals=terminals)
        if power_transformer_ends:
            for end in power_transformer_ends:
                self.add_end(end)

    def num_ends(self):
        """
        Get the number of `PowerTransformerEnd`s for this `PowerTransformer`.
        """
        return nlen(self._power_transformer_ends)

    @property
    def ends(self) -> Generator[PowerTransformerEnd, None, None]:
        """The `PowerTransformerEnd`s for this `PowerTransformer`."""
        return ngen(self._power_transformer_ends)

    def get_base_voltage(self, terminal: Terminal = None):
        if terminal is None:
            return self.base_voltage
        for end in self.ends:
            if end.terminal is terminal:
                return end.base_voltage
        else:
            return None

    def get_end_by_mrid(self, mrid: str) -> PowerTransformerEnd:
        """
        Get the `PowerTransformerEnd` for this `PowerTransformer` identified by `mrid`

        `mrid` the mRID of the required `PowerTransformerEnd`
        Returns The `PowerTransformerEnd` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._power_transformer_ends, mrid)

    def get_end_by_num(self, end_number: int):
        """
        Get the `PowerTransformerEnd` on this `PowerTransformer` by its `end_number`.

        `end_number` The `end_number` of the `PowerTransformerEnd` in relation to this `PowerTransformer`s VectorGroup.
        Returns The `PowerTransformerEnd` referred to by `end_number`
        Raises IndexError if no `PowerTransformerEnd` was found with end_number `end_number`.
        """
        if self._power_transformer_ends:
            for end in self._power_transformer_ends:
                if end.end_number == end_number:
                    return end
        raise IndexError(f"No TransformerEnd with end_number {end_number} was found in PowerTransformer {str(self)}")

    def add_end(self, end: PowerTransformerEnd) -> PowerTransformer:
        """
        Associate a `PowerTransformerEnd` with this `PowerTransformer`. If `end.end_number` == 0, the end will be assigned an end_number of
        `self.num_ends() + 1`.

        `end` the `PowerTransformerEnd` to associate with this `PowerTransformer`.
        Returns A reference to this `PowerTransformer` to allow fluent use.
        Raises `ValueError` if another `PowerTransformerEnd` with the same `mrid` already exists for this `PowerTransformer`.
        """
        if self._validate_end(end):
            return self

        if end.end_number == 0:
            end.end_number = self.num_ends() + 1

        self._power_transformer_ends = list() if self._power_transformer_ends is None else self._power_transformer_ends
        self._power_transformer_ends.append(end)
        self._power_transformer_ends.sort(key=lambda t: t.end_number)
        return self

    def remove_end(self, end: PowerTransformerEnd) -> PowerTransformer:
        """
        `end` the `PowerTransformerEnd` to disassociate from this `PowerTransformer`.
        Raises `ValueError` if `end` was not associated with this `PowerTransformer`.
        Returns A reference to this `PowerTransformer` to allow fluent use.
        """
        self._power_transformer_ends = safe_remove(self._power_transformer_ends, end)
        return self

    def clear_ends(self) -> PowerTransformer:
        """
        Clear all `PowerTransformerEnd`s.
        Returns A reference to this `PowerTransformer` to allow fluent use.
        """
        self._power_transformer_ends.clear()
        return self

    def _validate_end(self, end: PowerTransformerEnd) -> bool:
        """
        Validate an end against this `PowerTransformer`'s `PowerTransformerEnd`s.

        `end` The `PowerTransformerEnd` to validate.
        Returns True if `end` is already associated with this `PowerTransformer`, otherwise False.
        Raises `ValueError` if `end.power_transformer` is not this `PowerTransformer`, or if this `PowerTransformer` has a different `PowerTransformerEnd`
        with the same mRID.
        """
        if self._validate_reference(end, self.get_end_by_mrid, "A PowerTransformerEnd"):
            return True

        if self._validate_reference_by_sn(end.end_number, end, self.get_end_by_num, "A PowerTransformerEnd", "end_number"):
            return True

        require(end.power_transformer is self,
                lambda: f"PowerTransformerEnd {end} references another PowerTransformer {end.power_transformer}, expected {str(self)}.")
        return False



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations
from collections import defaultdict
from dataclassy import dataclass


from zepben.evolve.exceptions import PhaseException
from zepben.evolve.model.phasedirection import PhaseDirection
from zepben.evolve.model.cim.iec61970.base.wires.single_phase_kind import SinglePhaseKind, SINGLE_PHASE_KIND_VALUES

__all__ = ["phase", "direction", "pos_shift", "add", "setphs", "remove", "remove_all", "TracedPhases", "NominalPhasePath"]
CORE_MASKS = [0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000]
DIR_MASK = 0b11


PHASE_DIR_MAP = defaultdict(lambda: SinglePhaseKind.NONE)
PHASE_DIR_MAP[0b01] = SinglePhaseKind.A
PHASE_DIR_MAP[0b10] = SinglePhaseKind.A
PHASE_DIR_MAP[0b11] = SinglePhaseKind.A
PHASE_DIR_MAP[0b0100] = SinglePhaseKind.B
PHASE_DIR_MAP[0b1000] = SinglePhaseKind.B
PHASE_DIR_MAP[0b1100] = SinglePhaseKind.B
PHASE_DIR_MAP[0b010000] = SinglePhaseKind.C
PHASE_DIR_MAP[0b100000] = SinglePhaseKind.C
PHASE_DIR_MAP[0b110000] = SinglePhaseKind.C
PHASE_DIR_MAP[0b01000000] = SinglePhaseKind.N
PHASE_DIR_MAP[0b10000000] = SinglePhaseKind.N
PHASE_DIR_MAP[0b11000000] = SinglePhaseKind.N


def _valid_phase_check(nominal_phase):
    if nominal_phase not in SINGLE_PHASE_KIND_VALUES[1:7]:
        raise ValueError(f"INTERNAL ERROR: Phase {nominal_phase} is invalid. Must be one of {SINGLE_PHASE_KIND_VALUES[1:7]}.")


def phase(status: int, nominal_phase: SinglePhaseKind):
    core_val = (status >> (nominal_phase.mask_index * 8)) & 0xff
    return PHASE_DIR_MAP[core_val]


def direction(status: int, nominal_phase: SinglePhaseKind):
    dir_val = (status >> pos_shift(phase(status, nominal_phase), nominal_phase)) & DIR_MASK
    return PhaseDirection(dir_val)


def pos_shift(phs: SinglePhaseKind, nominal_phase: SinglePhaseKind):
    return (2 * max(phs.value - 1, 0)) + (nominal_phase.mask_index * 8)


def setphs(status: int, phs: SinglePhaseKind, direction: PhaseDirection, nominal_phs: SinglePhaseKind) -> int:
    return (status & ~CORE_MASKS[nominal_phs.mask_index]) | _shifted_value(direction, phs, nominal_phs)


def add(status: int, phs: SinglePhaseKind, direction: PhaseDirection, nominal_phs: SinglePhaseKind) -> int:
    return status | _shifted_value(direction, phs, nominal_phs)


def remove_all(status: int, nominal_phase: SinglePhaseKind) -> int:
    return status & ~CORE_MASKS[nominal_phase.mask_index]


def remove(status: int, phs: SinglePhaseKind, direction: PhaseDirection, nominal_phs: SinglePhaseKind) -> int:
    return status & ~_shifted_value(direction, phs, nominal_phs)


def _shifted_value(pd: PhaseDirection, spk: SinglePhaseKind, nom: SinglePhaseKind) -> int:
    return pd.value << pos_shift(spk, nom)


@dataclass(slots=True)
class NominalPhasePath(object):
    """
    Defines how a nominal phase is wired through a connectivity node between two terminals
    """

    from_phase: SinglePhaseKind
    """The nominal phase where the path comes from."""

    to_phase: SinglePhaseKind
    """The nominal phase where the path goes to."""


@dataclass(slots=True)
class TracedPhases(object):
    """
    Class that holds the traced phase statuses for the current and normal state of the network.
    Each byte in an int is used to store all possible phases and directions for a core.
    Each byte has 2 bits that represent the direction for a phase. If none of those bits are set the direction is equal to NONE.
    Use the figures below as a reference.
    <p>
    PhaseStatus:     |          integer          |
                     | byte | byte | byte | byte |
                     |Core 3|Core 2|Core 1|Core 0|
    <p>
    Core:            |                 byte                  |
                     |  2bits  |  2bits  |  2bits  |  2bits  |
    Phase:           |    N    |    C    |    B    |    A    |
    Direction:       |OUT | IN |OUT | IN |OUT | IN |OUT | IN |
    """
    _normal_status: int = 0
    _current_status: int = 0

    def __init__(self, normal_status: int = 0, current_status: int = 0):
        self._normal_status = normal_status
        self._current_status = current_status

    def __str__(self):
        s = []
        for phs in (SinglePhaseKind.A, SinglePhaseKind.B, SinglePhaseKind.C, SinglePhaseKind.N):
            pn = self.phase_normal(phs)
            pc = self.phase_current(phs)
            dn = self.direction_normal(phs)
            dc = self.direction_current(phs)
            s.append(f"phase {phs}: n: {pn.short_name}|{dn.short_name} c: {pc.short_name}|{dc.short_name}")
        return ", ".join(s)

    def phase_normal(self, nominal_phase: SinglePhaseKind):
        """
        Get the normal (nominal) phase for a core.
        `nominal_phase` The number of the core to check (between 0 - 4)
        Returns `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` for the core
        Raises `CoreException` if core is invalid.
        """
        _valid_phase_check(nominal_phase)
        return phase(self._normal_status, nominal_phase)

    def phase_current(self, nominal_phase: SinglePhaseKind):
        """
        Get the current (actual) phase for a core.
        `nominal_phase` The number of the core to check (between 0 - 4)
        Returns `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` for the core
        """
        _valid_phase_check(nominal_phase)
        return phase(self._current_status, nominal_phase)

    def direction_normal(self, nominal_phase: SinglePhaseKind):
        """
        Get the normal (nominal) direction for a core.
        `nominal_phase` The number of the core to check (between 0 - 4)
        Returns `zepben.phases.direction.Direction` for the core
        """
        _valid_phase_check(nominal_phase)
        return direction(self._normal_status, nominal_phase)

    def direction_current(self, nominal_phase: SinglePhaseKind):
        """
        Get the current (actual) direction for a core.
        `nominal_phase` The number of the core to check (between 0 - 4)
        Returns `zepben.phases.direction.Direction` for the core
        """
        _valid_phase_check(nominal_phase)
        return direction(self._current_status, nominal_phase)

    def add_normal(self, phs: SinglePhaseKind, nominal_phase: SinglePhaseKind, dir_: PhaseDirection):
        """
        Add a normal phase
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        Raises `zepben.phases.exceptions.PhaseException` if phases cross,
                 `zepben.phases.exceptions.CoreException` if `nominal_phase` is invalid.
        """
        _valid_phase_check(nominal_phase)
        if phs == SinglePhaseKind.NONE or dir_ == PhaseDirection.NONE:
            return False
        if self.phase_normal(nominal_phase) != SinglePhaseKind.NONE and phs != self.phase_normal(nominal_phase):
            raise PhaseException("Crossing phases")
        if self.direction_normal(nominal_phase).has(dir_):
            return False

        self._normal_status = add(self._normal_status, phs, dir_, nominal_phase)
        return True

    def add_current(self, phs: SinglePhaseKind, nominal_phase: SinglePhaseKind, dir_: PhaseDirection):
        """
        Add a current phase
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        Raises `zepben.phases.exceptions.PhaseException` if phases cross
        """
        _valid_phase_check(nominal_phase)
        if phs == SinglePhaseKind.NONE or dir_ == PhaseDirection.NONE:
            return False
        if self.phase_current(nominal_phase) != SinglePhaseKind.NONE and phs != self.phase_current(nominal_phase):
            raise PhaseException("Crossing phases")
        if self.direction_current(nominal_phase).has(dir_):
            return False

        self._current_status = add(self._current_status, phs, dir_, nominal_phase)
        return True

    def set_normal(self, phs: SinglePhaseKind, nominal_phase: SinglePhaseKind, dir_: PhaseDirection):
        """
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        """
        _valid_phase_check(nominal_phase)
        if phs == SinglePhaseKind.NONE or dir_ == PhaseDirection.NONE:
            self.remove_normal(self.phase_normal(nominal_phase), nominal_phase)
            return True

        if self.phase_normal(nominal_phase) == phs and self.direction_normal(nominal_phase) == dir_:
            return False

        self._normal_status = setphs(self._normal_status, phs, dir_, nominal_phase)
        return True

    def set_current(self, phs: SinglePhaseKind, dir_: PhaseDirection, nominal_phase: SinglePhaseKind):
        """
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        """
        _valid_phase_check(nominal_phase)
        if phs == SinglePhaseKind.NONE or dir_ == PhaseDirection.NONE:
            self.remove_current(self.phase_current(nominal_phase), nominal_phase)
            return True

        if self.phase_current(nominal_phase) == phs and self.direction_current(nominal_phase) == dir_:
            return False

        self._current_status = setphs(self._current_status, phs, dir_, nominal_phase)
        return True

    def remove_normal(self, phs: SinglePhaseKind, nominal_phase: SinglePhaseKind, dir_: PhaseDirection = None):
        """
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        """
        _valid_phase_check(nominal_phase)
        if phs != self.phase_normal(nominal_phase):
            return False

        if dir_ is not None:
            if not self.direction_normal(nominal_phase).has(dir_):
                return False
            self._normal_status = remove(self._normal_status, phs, dir_, nominal_phase)
        else:
            self._normal_status = remove_all(self._normal_status, nominal_phase)
            return True

    def remove_current(self, phs: SinglePhaseKind, nominal_phase: SinglePhaseKind, dir_: PhaseDirection = None):
        """
        `phs` The `zepben.protobuf.cim.iec61970.base.wires.SinglePhaseKind` to add.
        `nominal_phase` The core number this phase should be applied to
        `dir_` The direction of this phase relative to the location of the `zepben.evolve.iec61970.base.core.terminal.Terminal` to its feeder circuit breaker.
        Returns True if phase status was changed, False otherwise.
        """
        _valid_phase_check(nominal_phase)
        if phs != self.phase_current(nominal_phase):
            return False

        if dir_ is not None:
            if not self.direction_current(nominal_phase).has(dir_):
                return False
            self._current_status = remove(self._current_status, phs, dir_, nominal_phase)
        else:
            self._current_status = remove_all(self._current_status, nominal_phase)
            return True

    def copy(self):
        return TracedPhases()



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61970.base.wires.energy_connection import RegulatingCondEq
from zepben.evolve.model.cim.iec61970.base.wires.phase_shunt_connection_kind import PhaseShuntConnectionKind

__all__ = ["ShuntCompensator", "LinearShuntCompensator"]


class ShuntCompensator(RegulatingCondEq):
    """
    A shunt capacitor or reactor or switchable bank of shunt capacitors or reactors. A section of a shunt compensator
    is an individual capacitor or reactor.  A negative value for reactivePerSection indicates that the compensator is
    a reactor. ShuntCompensator is a single terminal device.  Ground is implied.
    """
    grounded: bool = False
    """Used for Yn and Zn connections. True if the neutral is solidly grounded. nom_u : The voltage at which the nominal reactive power may be calculated. 
    This should normally be within 10% of the voltage at which the capacitor is connected to the network."""

    nom_u: int = 0
    """The voltage at which the nominal reactive power may be calculated. This should normally be within 10% of the voltage at which the capacitor is connected 
    to the network."""

    phase_connection: PhaseShuntConnectionKind = PhaseShuntConnectionKind.UNKNOWN
    """The type of phase connection, such as wye or delta."""

    sections: float = 0.0
    """
    Shunt compensator sections in use. Starting value for steady state solution. Non integer values are allowed to support continuous variables. The 
    reasons for continuous value are to support study cases where no discrete shunt compensator's has yet been designed, a solutions where a narrow voltage 
    band force the sections to oscillate or accommodate for a continuous solution as input. 
    
    For `LinearShuntCompensator` the value shall be between zero and `ShuntCompensator.maximumSections`. At value zero the shunt compensator conductance and 
    admittance is zero. Linear interpolation of conductance and admittance between the previous and next integer section is applied in case of non-integer 
    values. 
    
    For `NonlinearShuntCompensator`s shall only be set to one of the NonlinearShuntCompensatorPoint.sectionNumber. There is no interpolation between 
    NonlinearShuntCompensatorPoint-s.
    """


class LinearShuntCompensator(ShuntCompensator):
    """A linear shunt compensator has banks or sections with equal admittance values."""

    b0_per_section: float = 0.0
    """Zero sequence shunt (charging) susceptance per section"""

    b_per_section: float = 0.0
    """Positive sequence shunt (charging) susceptance per section"""

    g0_per_section: float = 0.0
    """Zero sequence shunt (charging) conductance per section"""

    g_per_section: float = 0.0
    """Positive sequence shunt (charging) conductance per section"""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61968.common.location import Location
from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject
from zepben.evolve.util import get_by_mrid, nlen, ngen, safe_remove

__all__ = ["Asset", "AssetContainer"]


class Asset(IdentifiedObject):
    """
    Tangible resource of the utility, including power system equipment, various end devices, cabinets, buildings,
    etc. For electrical network equipment, the role of the asset is defined through PowerSystemResource and its
    subclasses, defined mainly in the Wires model (refer to IEC61970-301 and model package IEC61970::Wires). Asset
    description places emphasis on the physical characteristics of the equipment fulfilling that role.
    """

    location: Optional[Location] = None
    """`zepben.evolve.cim.iec61968.common.location.Location` of this asset"""

    _organisation_roles: Optional[List[AssetOrganisationRole]] = None

    def __init__(self, organisation_roles: List[AssetOrganisationRole] = None):
        if organisation_roles:
            for role in organisation_roles:
                self.add_organisation_role(role)

    def num_organisation_roles(self) -> int:
        """
        Get the number of `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole`s associated with this `Asset`.
        """
        return nlen(self._organisation_roles)

    @property
    def organisation_roles(self) -> Generator[AssetOrganisationRole, None, None]:
        """
        The `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole`s of this `Asset`.
        """
        return ngen(self._organisation_roles)

    def get_organisation_role(self, mrid: str) -> AssetOrganisationRole:
        """
        Get the `AssetOrganisationRole` for this asset identified by `mrid`.

        `mrid` the mRID of the required `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole`
        Returns The `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole` with the specified `mrid`.
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._organisation_roles, mrid)

    def add_organisation_role(self, role: AssetOrganisationRole) -> Asset:
        """
        `role` The `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole` to
        associate with this `Asset`.
        Returns A reference to this `Asset` to allow fluent use.
        Raises `ValueError` if another `AssetOrganisationRole` with the same `mrid` already exists in this `Asset`
        """
        if self._validate_reference(role, self.get_organisation_role, "An AssetOrganisationRole"):
            return self

        self._organisation_roles = list() if self._organisation_roles is None else self._organisation_roles
        self._organisation_roles.append(role)
        return self

    def remove_organisation_role(self, role: AssetOrganisationRole) -> Asset:
        """
        Disassociate an `AssetOrganisationRole` from this `Asset`.

        `role` the `zepben.evolve.cim.iec61968.assets.asset_organisation_role.AssetOrganisationRole` to
        disassociate with this `Asset`.
        Raises `ValueError` if `role` was not associated with this `Asset`.
        Returns A reference to this `Asset` to allow fluent use.
        """
        self._organisation_roles = safe_remove(self._organisation_roles, role)
        return self

    def clear_organisation_roles(self) -> Asset:
        """
        Clear all organisation roles.
        Returns self
        """
        self._organisation_roles = None
        return self


class AssetContainer(Asset):
    """
    Asset that is aggregation of other assets such as conductors, transformers, switchgear, land, fences, buildings,
    equipment, vehicles, etc.
    """
    pass



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject

__all__ = ["AssetInfo"]


class AssetInfo(IdentifiedObject):
    """
    Set of attributes of an asset, representing typical datasheet information of a physical device that can be
    instantiated and shared in different data exchange contexts:
        - as attributes of an asset instance (installed or in stock)
        - as attributes of an asset model (product by a manufacturer)
        - as attributes of a type asset (generic type of an asset as used in designs/extension planning).
    """
    pass




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pytest
from zepben.evolve.services.network.tracing.traversals.tracing import Traversal
from zepben.evolve.services.network.tracing.traversals import BranchRecursiveTraversal
from zepben.evolve.services.network.tracing.traversals.queue import FifoQueue, LifoQueue
from typing import List, Optional, Set


async def validate_run(t: Traversal, visit_order: List[int], expected_order: List[int], can_stop_on_start=True, check_visited=True):
    # clean slate each run
    t.reset()
    visit_order.clear()
    await t.trace(can_stop_on_start_item=can_stop_on_start)
    for i, x in enumerate(expected_order):
        assert visit_order[i] == x
        if check_visited:
            assert t.tracker.has_visited(x)


async def _validate_can_stop(t: Traversal, visit_order: List[int], expected_order: List[int], stop_count=None, check_visited=True):
    await validate_run(t, visit_order=visit_order, expected_order=expected_order, can_stop_on_start=False, check_visited=check_visited)
    if stop_count is not None:
        assert stop_count == len(visit_order)
    await validate_run(t, visit_order=visit_order, expected_order=[expected_order[0]], can_stop_on_start=True, check_visited=check_visited)
    if stop_count is not None:
        assert stop_count == len(visit_order) - 1


def queue_next(item: int, exclude=None):
    return filter(lambda x: x > 0 and x not in exclude, [item - 2, item - 1, item + 1, item + 2])


class TestTracing(object):

    @pytest.mark.asyncio
    async def test_breadth_first(self):
        expected_order = [1, 2, 3, 4, 5, 6, 7]
        visit_order = []

        async def cond(i):
            return i >= 6

        async def action(i, s):
            visit_order.append(i)

        t = Traversal(queue_next=queue_next, start_item=1, process_queue=FifoQueue(), stop_conditions=[cond], step_actions=[action])

        await validate_run(t, can_stop_on_start=True, visit_order=visit_order, expected_order=expected_order)

    @pytest.mark.asyncio
    async def test_depth_first(self):
        expected_order = [1, 3, 5, 7, 6, 4, 2]
        visit_order = []

        async def cond(i):
            return i >= 6

        async def action(i, s):
            visit_order.append(i)

        t = Traversal(queue_next=queue_next, start_item=1, process_queue=LifoQueue(), stop_conditions=[cond], step_actions=[action])

        await validate_run(t, can_stop_on_start=True, visit_order=visit_order, expected_order=expected_order)

    @pytest.mark.asyncio
    async def test_can_stop_on_start_item(self):
        async def cond1(i):
            return i >= 0

        async def cond2(i):
            return i >= 6

        visit_order = []

        async def action(i, s):
            visit_order.append(i)

        t = Traversal(queue_next=queue_next, start_item=1, process_queue=FifoQueue(), stop_conditions=[cond1, cond2], step_actions=[action])
        await _validate_can_stop(t, visit_order=visit_order, expected_order=[1, 2, 3])
        t = Traversal(queue_next=queue_next, start_item=1, process_queue=LifoQueue(), stop_conditions=[cond1, cond2], step_actions=[action])
        await _validate_can_stop(t, visit_order=visit_order, expected_order=[1, 3, 2])

    @pytest.mark.asyncio
    async def test_stopping_to_step(self):
        visited = set()
        stopping_on = set()

        async def cond(i):
            return i >= 3

        async def action(i, s):
            visited.add(i)
            if s:
                stopping_on.add(i)

        t = Traversal(queue_next=lambda i, exc: [i + 1, i + 2], start_item=1, process_queue=LifoQueue(), stop_conditions=[cond], step_actions=[action])

        await t.trace(can_stop_on_start_item=True)
        for x in range(1, 4):
            assert x in visited
        for x in range(3, 4):
            assert x in stopping_on


def queue_next_br(item: int, traversal: BranchRecursiveTraversal, exclude: Optional[Set[int]] = None):
    if item == 0:
        branch = traversal.create_branch()
        branch.start_item = 1
        traversal.branch_queue.put(branch)
        branch = traversal.create_branch()
        branch.start_item = 3
        traversal.branch_queue.put(branch)
    elif item == 1 or item == 3:
        if traversal.tracker.has_visited(2):
            traversal.process_queue.put(0)
        else:
            traversal.process_queue.put(2)
    elif item == 2:
        if traversal.tracker.has_visited(1):
            traversal.process_queue.put(3)
        elif traversal.tracker.has_visited(3):
            traversal.process_queue.put(1)


class TestBranchRecursiveTraversal(object):

    @pytest.mark.asyncio
    async def test_simple(self):
        visited = list()

        async def action(i, s):
            visited.append(i)

        self.stop_count = 0

        async def cond(i):
            self.stop_count += 1
            return False

        t = BranchRecursiveTraversal(start_item=0, queue_next=queue_next_br, process_queue=LifoQueue(), branch_queue=FifoQueue(), step_actions=[action],
                                     stop_conditions=[cond])
        await validate_run(t, visited, [0, 1, 2, 3, 3, 2, 1], check_visited=False)
        assert self.stop_count == len(visited)

    @pytest.mark.asyncio
    async def test_stop_first_asset(self):
        visited = list()

        async def action(i, s):
            visited.append(i)

        self.stop_count = 0

        async def cond1(i):
            self.stop_count += 1
            return False

        async def cond2(i):
            return i == 0

        t = BranchRecursiveTraversal(start_item=0, queue_next=queue_next_br, process_queue=LifoQueue(), branch_queue=FifoQueue(), step_actions=[action],
                                     stop_conditions=[cond1, cond2])
        await _validate_can_stop(t, visited, [0, 1, 2, 3, 3, 2, 1], check_visited=False)



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from zepben.evolve import PowerTransformer, BaseVoltage, Terminal, PowerTransformerEnd


class TestPowerTransformer(object):

    def test_powertransformer_voltages(self):
        bv11k = BaseVoltage(nominal_voltage=11000)
        pt = PowerTransformer(base_voltage=bv11k)
        t1 = Terminal(conducting_equipment=pt)
        pte = PowerTransformerEnd(power_transformer=pt, base_voltage=bv11k, terminal=t1)
        pt.add_end(pte)
        assert pt.get_base_voltage(t1) is pte.base_voltage is t1.base_voltage



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import contextlib
import grpc
import requests
import json
from jose import jwt
from datetime import datetime
from zepben.evolve.streaming.exceptions import AuthException

__all__ = ["connect", "connect_async"]
_AUTH_HEADER_KEY = 'authorization'


class AuthTokenPlugin(grpc.AuthMetadataPlugin):

    def __init__(self, host, conf_address, client_id, client_secret):
        self.host = host
        self.conf_address = conf_address
        self.client_id = client_id
        self.client_secret = client_secret
        self.token = ""
        self.token_expiry = 0
        self._refresh_token()

    def __call__(self, context, callback):
        if datetime.utcnow().timestamp() > self.token_expiry:
            self._refresh_token()
        callback(((_AUTH_HEADER_KEY, self.token),), None)

    def _refresh_token(self):
        parts = get_token(self.host, self.conf_address, self.client_id, self.client_secret)
        self.token = f"{parts['token_type']} {parts['access_token']}"
        self.token_expiry = jwt.get_unverified_claims(parts['access_token'])['exp']


def get_token(addr, conf_address, client_id, client_secret):
    # Get the configuration TODO: this probably needs to be OAuth2 compliant or something
    with requests.session() as session:
        with session.get(conf_address) as resp:
            result = json.loads(resp.text)
            domain = result["dom"]
            aud = result["aud"]
        with session.post(domain, data={'client_id': client_id, 'client_secret': client_secret, 'audience': aud, 'grant_type': 'client_credentials'}) as resp:
            token = json.loads(resp.text)
    if 'error' in token:
        raise AuthException(f"{token['error']}: {token['error_description']}")
    return token


def _conn(host: str = "localhost", rpc_port: int = 50051, conf_address: str = "http://localhost/auth", client_id: str = None,
          client_secret: str = None, pkey=None, cert=None, ca=None):
    """
    `host` The host to connect to.
    `rpc_port` The gRPC port for host.
    `conf_address` The complete address for the auth configuration endpoint.
    `client_id` Your client id for your OAuth Auth provider.
    `client_secret` Corresponding client secret.
    `pkey` Private key for client authentication
    `cert` Corresponding signed certificate. CN must reflect your hosts FQDN, and must be signed by the servers
                 CA.
    `ca` CA trust for the server.
    `secure_conf` Whether the server hosting configuration is secured (https)
    Returns A gRPC channel
    """
    # TODO: make this more robust so it can handle SSL without client verification
    if pkey and cert and client_id and client_secret:
        call_credentials = grpc.metadata_call_credentials(AuthTokenPlugin(host, conf_address, client_id, client_secret))
        # Channel credential will be valid for the entire channel
        channel_credentials = grpc.ssl_channel_credentials(ca, pkey, cert)
        # Combining channel credentials and call credentials together
        composite_credentials = grpc.composite_channel_credentials(
            channel_credentials,
            call_credentials,
        )
        channel = grpc.secure_channel(f"{host}:{rpc_port}", composite_credentials)
    else:
        channel = grpc.insecure_channel(f"{host}:{rpc_port}")

    return channel


@contextlib.contextmanager
def connect(host: str = "localhost",
            rpc_port: int = 50051,
            conf_address: str = "http://localhost/auth",
            client_id: str = None,
            client_secret: str = None,
            pkey=None,
            cert=None,
            ca=None):
    """
    Usage:
        with connect(args) as channel:

    `host` The host to connect to.
    `rpc_port` The gRPC port for host.
    `conf_address` The complete address for the auth configuration endpoint.
    `client_id` Your client id for your OAuth Auth provider.
    `client_secret` Corresponding client secret.
    `pkey` Private key for client authentication
    `cert` Corresponding signed certificate. CN must reflect your hosts FQDN, and must be signed by the servers
                 CA.
    `ca` CA trust for the server.
    Returns A gRPC channel
    """
    yield _conn(host, rpc_port, conf_address, client_id, client_secret, pkey, cert, ca)


@contextlib.asynccontextmanager
async def connect_async(host: str = "localhost",
                        rpc_port: int = 50051,
                        conf_address: str = "http://localhost/auth",
                        client_id: str = None,
                        client_secret: str = None,
                        pkey=None,
                        cert=None,
                        ca=None):
    """
    Usage:
        async with connect_async(args) as channel:

    `host` The host to connect to.
    `rpc_port` The gRPC port for host.
    `conf_address` The complete address for the auth configuration endpoint.
    `client_id` Your client id for your OAuth Auth provider.
    `client_secret` Corresponding client secret.
    `pkey` Private key for client authentication
    `cert` Corresponding signed certificate. CN must reflect your hosts FQDN, and must be signed by the servers
                 CA.
    `ca` CA trust for the server.
    Returns A gRPC channel
    """
    yield _conn(host, rpc_port, conf_address, client_id, client_secret, pkey, cert, ca)



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

test_deps = ["pytest", "pytest-asyncio", "hypothesis<6"]
setup(
    name="zepben.evolve",
    version="0.21.0b1",
    description="Python SDK for interacting with the Evolve platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zepben/evolve-sdk-python",
    author="Kurt Greaves",
    author_email="kurt.greaves@zepben.com",
    license="MPL 2.0",
    classifiers=[
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src"),
    python_requires='>=3.7',
    install_requires=[
        "protobuf",
        "requests",
        "zepben.protobuf",
        "python-jose-cryptodome",
        "dataclassy"
    ],
    extras_require={
        "test": test_deps,
    }
)



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

import pytest
from zepben.evolve import SetPhases, phase_log, PhaseDirection, SinglePhaseKind
from test.util import get_terminal, check_phases

A = SinglePhaseKind.A
B = SinglePhaseKind.B
C = SinglePhaseKind.C
N = SinglePhaseKind.N
SPK_NONE = SinglePhaseKind.NONE

OUT = PhaseDirection.OUT
IN = PhaseDirection.IN
BOTH = PhaseDirection.BOTH
NONE = PhaseDirection.NONE


class TestSetPhase(object):
    @pytest.mark.asyncio
    async def test_set_phase(self, network1):
        assert network1 is not None
        set_phases = SetPhases()
        await set_phases.run(network1)
        assert len(network1.energy_sources) == 1
        assert len(network1.energy_consumers) == 1

        check_phases(get_terminal(network1, "default-acls-1", 0), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network1, "default-acls-1", 1), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network1, "acls-1", 1), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network1, "acls-1", 0), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network1, "default-es", 0), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network1, "default-cb", 0), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network1, "default-cb", 1), [SPK_NONE, SPK_NONE, SPK_NONE, SPK_NONE], [NONE, NONE, NONE, NONE])
        check_phases(get_terminal(network1, "trafo-1", 0), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network1, "trafo-1", 1), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network1, "ec-1", 0), [A, B, C, N], [IN, IN, IN, IN])

    @pytest.mark.asyncio
    async def test_set_phase_multi_branch(self, network2):
        assert network2 is not None
        set_phases = SetPhases()
        await set_phases.run(network2)
        breakers = []
        for br in network2.breakers.values():
            try:
                if br.is_substation_breaker():
                    breakers.append(br)
            except:
                pass

        await phase_log(breakers)
        check_phases(get_terminal(network2, "acls0", 0), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network2, "acls0", 1), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network2, "es", 0), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network2, "acls1", 0), [A, B, C, N], [IN, IN, IN, IN])
        check_phases(get_terminal(network2, "acls1", 1), [A, B, C, N], [OUT, OUT, OUT, OUT])
        check_phases(get_terminal(network2, "acls4", 0), [A, B], [IN, IN])
        check_phases(get_terminal(network2, "acls4", 1), [A, B], [OUT, OUT])
        check_phases(get_terminal(network2, "acls7", 1), [A, B], [OUT, OUT])
        check_phases(get_terminal(network2, "acls8", 0), [B, C], [OUT, OUT])
        check_phases(get_terminal(network2, "acls9", 0), [B, C], [OUT, OUT])
        check_phases(get_terminal(network2, "acls9", 1), [B, C], [IN, IN])
        check_phases(get_terminal(network2, "br0", 0), [A, B], [IN, IN])
        check_phases(get_terminal(network2, "br0", 1), [B, C], [IN, IN])
        check_phases(get_terminal(network2, "junc7", 0), [B], [IN])





#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.protobuf.cim.iec61968.assetinfo.CableInfo_pb2 import CableInfo
from zepben.protobuf.cim.iec61968.assetinfo.OverheadWireInfo_pb2 import OverheadWireInfo
from zepben.protobuf.cim.iec61968.assets.AssetOwner_pb2 import AssetOwner
from zepben.protobuf.cim.iec61968.assets.Pole_pb2 import Pole
from zepben.protobuf.cim.iec61968.assets.Streetlight_pb2 import Streetlight
from zepben.protobuf.cim.iec61968.common.Location_pb2 import Location
from zepben.protobuf.cim.iec61968.common.Organisation_pb2 import Organisation
from zepben.protobuf.cim.iec61968.metering.Meter_pb2 import Meter
from zepben.protobuf.cim.iec61968.metering.UsagePoint_pb2 import UsagePoint
from zepben.protobuf.cim.iec61968.operations.OperationalRestriction_pb2 import OperationalRestriction
from zepben.protobuf.cim.iec61970.base.auxiliaryequipment.FaultIndicator_pb2 import FaultIndicator
from zepben.protobuf.cim.iec61970.base.core.BaseVoltage_pb2 import BaseVoltage
from zepben.protobuf.cim.iec61970.base.core.ConnectivityNode_pb2 import ConnectivityNode
from zepben.protobuf.cim.iec61970.base.core.Feeder_pb2 import Feeder
from zepben.protobuf.cim.iec61970.base.core.GeographicalRegion_pb2 import GeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.Site_pb2 import Site
from zepben.protobuf.cim.iec61970.base.core.SubGeographicalRegion_pb2 import SubGeographicalRegion
from zepben.protobuf.cim.iec61970.base.core.Substation_pb2 import Substation
from zepben.protobuf.cim.iec61970.base.core.Terminal_pb2 import Terminal
from zepben.protobuf.cim.iec61970.base.wires.AcLineSegment_pb2 import AcLineSegment
from zepben.protobuf.cim.iec61970.base.wires.Breaker_pb2 import Breaker
from zepben.protobuf.cim.iec61970.base.wires.Disconnector_pb2 import Disconnector
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumer_pb2 import EnergyConsumer
from zepben.protobuf.cim.iec61970.base.wires.EnergyConsumerPhase_pb2 import EnergyConsumerPhase
from zepben.protobuf.cim.iec61970.base.wires.EnergySource_pb2 import EnergySource
from zepben.protobuf.cim.iec61970.base.wires.EnergySourcePhase_pb2 import EnergySourcePhase
from zepben.protobuf.cim.iec61970.base.wires.Fuse_pb2 import Fuse
from zepben.protobuf.cim.iec61970.base.wires.Jumper_pb2 import Jumper
from zepben.protobuf.cim.iec61970.base.wires.Junction_pb2 import Junction
from zepben.protobuf.cim.iec61970.base.wires.LinearShuntCompensator_pb2 import LinearShuntCompensator
from zepben.protobuf.cim.iec61970.base.wires.PerLengthSequenceImpedance_pb2 import PerLengthSequenceImpedance
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformer_pb2 import PowerTransformer
from zepben.protobuf.cim.iec61970.base.wires.PowerTransformerEnd_pb2 import PowerTransformerEnd
from zepben.protobuf.cim.iec61970.base.wires.RatioTapChanger_pb2 import RatioTapChanger
from zepben.protobuf.cim.iec61970.base.wires.Recloser_pb2 import Recloser
from zepben.protobuf.cim.iec61970.base.diagramlayout.Diagram_pb2 import Diagram
from zepben.protobuf.cim.iec61970.base.diagramlayout.DiagramObject_pb2 import DiagramObject
from zepben.protobuf.cim.iec61968.customers.Customer_pb2 import Customer
from zepben.protobuf.cim.iec61968.customers.CustomerAgreement_pb2 import CustomerAgreement
from zepben.protobuf.cim.iec61968.customers.PricingStructure_pb2 import PricingStructure
from zepben.protobuf.cim.iec61968.customers.Tariff_pb2 import Tariff
from zepben.protobuf.cim.iec61968.common.Organisation_pb2 import Organisation
from zepben.protobuf.dp.dp_requests_pb2 import *
from zepben.protobuf.cp.cp_requests_pb2 import *
from zepben.protobuf.cp.cp_requests_pb2 import CreateOrganisationRequest as CreateCustomerOrganisationRequest
from zepben.protobuf.np.np_requests_pb2 import *

network_rpc_map = {
    CableInfo: ('CreateCableInfo', CreateCableInfoRequest),
    OverheadWireInfo: ('CreateOverheadWireInfo', CreateOverheadWireInfoRequest),
    AssetOwner: ('CreateAssetOwner', CreateAssetOwnerRequest),
    Pole: ('CreatePole', CreatePoleRequest),
    Streetlight: ('CreateStreetlight', CreateStreetlightRequest),
    Location: ('CreateLocation', CreateLocationRequest),
    Organisation: ('CreateOrganisation', CreateOrganisationRequest),
    Meter: ('CreateMeter', CreateMeterRequest),
    UsagePoint: ('CreateUsagePoint', CreateUsagePointRequest),
    OperationalRestriction: ('CreateOperationalRestriction', CreateOperationalRestrictionRequest),
    FaultIndicator: ('CreateFaultIndicator', CreateFaultIndicatorRequest),
    BaseVoltage: ('CreateBaseVoltage', CreateBaseVoltageRequest),
    ConnectivityNode: ('CreateConnectivityNode', CreateConnectivityNodeRequest),
    Feeder: ('CreateFeeder', CreateFeederRequest),
    GeographicalRegion: ('CreateGeographicalRegion', CreateGeographicalRegionRequest),
    Site: ('CreateSite', CreateSiteRequest),
    SubGeographicalRegion: ('CreateSubGeographicalRegion', CreateSubGeographicalRegionRequest),
    Substation: ('CreateSubstation', CreateSubstationRequest),
    Terminal: ('CreateTerminal', CreateTerminalRequest),
    AcLineSegment: ('CreateAcLineSegment', CreateAcLineSegmentRequest),
    EnergyConsumer: ('CreateEnergyConsumer', CreateEnergyConsumerRequest),
    Disconnector: ('CreateDisconnector', CreateDisconnectorRequest),
    Breaker: ('CreateBreaker', CreateBreakerRequest),
    EnergyConsumerPhase: ('CreateEnergyConsumerPhase', CreateEnergyConsumerPhaseRequest),
    EnergySource: ('CreateEnergySource', CreateEnergySourceRequest),
    EnergySourcePhase: ('CreateEnergySourcePhase', CreateEnergySourcePhaseRequest),
    Fuse: ('CreateFuse', CreateFuseRequest),
    Jumper: ('CreateJumper', CreateJumperRequest),
    Junction: ('CreateJunction', CreateJunctionRequest),
    LinearShuntCompensator: ('CreateLinearShuntCompensator', CreateLinearShuntCompensatorRequest),
    PerLengthSequenceImpedance: ('CreatePerLengthSequenceImpedance', CreatePerLengthSequenceImpedanceRequest),
    PowerTransformer: ('CreatePowerTransformer', CreatePowerTransformerRequest),
    PowerTransformerEnd: ('CreatePowerTransformerEnd', CreatePowerTransformerEndRequest),
    RatioTapChanger: ('CreateRatioTapChanger', CreateRatioTapChangerRequest),
    Recloser: ('CreateRecloser', CreateRecloserRequest),
}

diagram_rpc_map = {
    Diagram: ('CreateDiagram', CreateDiagramRequest),
    DiagramObject: ('CreateDiagramObject', CreateDiagramObjectRequest),
}

customer_rpc_map = {
    Customer: ('CreateCustomer', CreateCustomerRequest),
    CustomerAgreement: ('CreateCustomerAgreement', CreateCustomerAgreementRequest),
    PricingStructure: ('CreatePricingStructure', CreatePricingStructureRequest),
    Tariff: ('CreateTariff', CreateTariffRequest),
    Organisation: ('CreateOrganisation', CreateCustomerOrganisationRequest),
}


#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["UnitSymbol", "unit_symbol_from_id", "unit_symbol_from_cim_name"]


def unit_symbol_from_cim_name(value: str):
    return _unitsymbol_by_cim_name[value]


def unit_symbol_from_id(value: int):
    return _unitsymbol_members_by_id[value]


class UnitSymbol(Enum):
    """
    The derived units defined for usage in the CIM. In some cases, the derived unit is equal to an SI unit. Whenever possible, the standard derived symbol is
    used instead of the formula for the derived unit. For example, the unit symbol Farad is defined as “F” instead of “CPerV”. In cases where a standard
    symbol does not exist for a derived unit, the formula for the unit is used as the unit symbol. For example, density does not have a standard symbol and
    so it is represented as “kgPerm3”. With the exception of the “kg”, which is an SI unit, the unit symbols do not contain multipliers and therefore
    represent the base derived unit to which a multiplier can be applied as a whole. Every unit symbol is treated as an unparseable text as if it were a
    single-letter symbol. The meaning of each unit symbol is defined by the accompanying descriptive text and not by the text contents of the unit symbol. To
    allow the widest possible range of serializations without requiring special character handling, several substitutions are made which deviate from the
    format described in IEC 80000-1. The division symbol “/” is replaced by the letters “Per”. Exponents are written in plain text after the unit as “m3”
    instead of being formatted as “m” with a superscript of 3 or introducing a symbol as in “m^3”. The degree symbol “°” is replaced with the letters “deg”.
    Any clarification of the meaning for a substitution is included in the description for the unit symbol. Non-SI units are included in list of unit symbols
    to allow sources of data to be correctly labelled with their non-SI units (for example, a GPS sensor that is reporting numbers that represent feet
    instead of meters). This allows software to use the unit symbol information correctly convert and scale the raw data of those sources into SI-based
    units. The integer values are used for harmonization with IEC 61850.
    """

    NONE = (0, "none")
    """Dimension less quantity, e.g. count, per unit, etc."""

    METRES = (1, "m")
    """Length in metres."""

    KG = (2, "kg")
    """Mass in kilograms. Note: multiplier “k” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    SECONDS = (3, "s")
    """Time in seconds."""

    A = (4, "A")
    """Current in amperes."""

    K = (5, "K")
    """Temperature in kelvins."""

    MOL = (6, "mol")
    """Amount of substance in moles."""

    CD = (7, "cd")
    """Luminous intensity in candelas."""

    DEG = (8, "deg")
    """Plane angle in degrees."""

    RAD = (9, "rad")
    """Plane angle in radians (m/m)."""

    SR = (10, "sr")
    """Solid angle in steradians (m2/m2)."""

    GY = (11, "Gy")
    """Absorbed dose in grays (J/kg)."""

    BQ = (12, "Bq")
    """Radioactivity in becquerels (1/s)."""

    DEGC = (13, "degC")
    """Relative temperature in degrees Celsius.In the SI unit system the symbol is °C. Electric charge is measured in coulomb that has the unit symbol C. 
    To distinguish degree Celsius from coulomb the symbol used in the UML is degC. The reason for not using °C is that the special character ° is difficult to 
    manage in software."""" """

    SV = (14, "Sv")
    """Dose equivalent in sieverts (J/kg)."""

    F = (15, "F")
    """Electric capacitance in farads (C/V)."""

    C = (16, "C")
    """Electric charge in coulombs (A·s)."""

    SIEMENS = (17, "S")
    """Conductance in siemens."""

    HENRYS = (18, "H")
    """Electric inductance in henrys (Wb/A)."""

    V = (19, "V")
    """Electric potential in volts (W/A)."""

    OHM = (20, "ohm")
    """Electric resistance in ohms (V/A)."""

    J = (21, "J")
    """Energy in joules (N·m = C·V = W·s)."""

    N = (22, "N")
    """Force in newtons (kg·m/s²)."""

    HZ = (23, "Hz")
    """Frequency in hertz (1/s)."""

    LX = (24, "lx")
    """Illuminance in lux (lm/m²)."""

    LM = (25, "lm")
    """Luminous flux in lumens (cd·sr)."""

    WB = (26, "Wb")
    """Magnetic flux in webers (V·s)."""

    T = (27, "T")
    """Magnetic flux density in teslas (Wb/m2)."""

    W = (28, "W")
    """Real power in watts (J/s). Electrical power may have real and reactive components. The real portion of electrical power (I²R or VIcos(phi)), 
    is expressed in Watts. See also apparent power and reactive power."""

    PA = (29, "Pa")
    """Pressure in pascals (N/m²). Note: the absolute or relative measurement of pressure is implied with this entry. See below for more explicit forms."""

    M2 = (30, "m2")
    """Area in square metres (m²)."""

    M3 = (31, "m3")
    """Volume in cubic metres (m³)."""

    MPERS = (32, "mPers")
    """Velocity in metres per second (m/s)."""

    MPERS2 = (33, "mPers2")
    """Acceleration in metres per second squared (m/s²)."""

    M3PERS = (34, "m3Pers")
    """Volumetric flow rate in cubic metres per second (m³/s)."""

    MPERM3 = (35, "mPerm3")
    """Fuel efficiency in metres per cubic metres (m/m³)."""

    KGM = (36, "kgm")
    """Moment of mass in kilogram metres (kg·m) (first moment of mass). 
    Note: multiplier “k” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    KGPERM3 = (37, "kgPerm3")
    """Density in kilogram/cubic metres (kg/m³). Note: multiplier “k” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    M2PERS = (38, "m2Pers")
    """Viscosity in square metres / second (m²/s)."""

    WPERMK = (39, "WPermK")
    """Thermal conductivity in watt/metres kelvin."""

    JPERK = (40, "JPerK")
    """Heat capacity in joules/kelvin."""

    PPM = (41, "ppm")
    """Concentration in parts per million."""

    ROTPERS = (42, "rotPers")
    """Rotations per second (1/s). See also Hz (1/s)."""

    RADPERS = (43, "radPers")
    """Angular velocity in radians per second (rad/s)."""

    WPERM2 = (44, "WPerm2")
    """Heat flux density, irradiance, watts per square metre."""

    JPERM2 = (45, "JPerm2")
    """Insulation energy density, joules per square metre or watt second per square metre."""

    SPERM = (46, "SPerm")
    """Conductance per length (F/m)."""

    KPERS = (47, "KPers")
    """Temperature change rate in kelvins per second."""

    PAPERS = (48, "PaPers")
    """Pressure change rate in pascals per second."""

    JPERKGK = (49, "JPerkgK")
    """Specific heat capacity, specific entropy, joules per kilogram Kelvin."""

    VA = (50, "VA")
    """Apparent power in volt amperes. See also real power and reactive power."""

    VAR = (51, "VAr")
    """Reactive power in volt amperes reactive. The “reactive” or “imaginary” component of electrical power (VIsin(phi)). (See also real power and apparent power). 
    Note: Different meter designs use different methods to arrive at their results. Some meters may compute reactive power as an arithmetic value, while others 
    compute the value vectorially. The data consumer should determine the method in use and the suitability of the measurement for the intended purpose."""

    COSPHI = (52, "cosPhi")
    """Power factor, dimensionless. 
    Note 1: This definition of power factor only holds for balanced systems. See the alternative definition under code 153. 
    Note 2 : Beware of differing sign conventions in use between the IEC and EEI. It is assumed that the data consumer understands the type of meter in use 
    and the sign convention in use by the utility."""

    VS = (53, "Vs")
    """Volt seconds (Ws/A)."""

    V2 = (54, "V2")
    """Volt squared (W²/A²)."""

    AS = (55, "As")
    """Ampere seconds (A·s)."""

    A2 = (56, "A2")
    """Amperes squared (A²)."""

    A2S = (57, "A2s")
    """Ampere squared time in square amperes (A²s)."""

    VAH = (58, "VAh")
    """Apparent energy in volt ampere hours."""

    WH = (59, "Wh")
    """Real energy in watt hours."""

    VARH = (60, "VArh")
    """Reactive energy in volt ampere reactive hours."""

    VPERHZ = (61, "VPerHz")
    """Magnetic flux in volt per hertz."""

    HZPERS = (62, "HzPers")
    """Rate of change of frequency in hertz per second."""

    CHARACTER = (63, "character")
    """Number of characters."""

    CHARPERS = (64, "charPers")
    """Data rate (baud) in characters per second."""

    KGM2 = (65, "kgm2")
    """Moment of mass in kilogram square metres (kg·m²) (Second moment of mass, commonly called the moment of inertia). Note: multiplier “k” is included in 
    this unit symbol for compatibility with IEC 61850-7-3."""

    DB = (66, "dB")
    """Sound pressure level in decibels. Note: multiplier “d” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    WPERS = (67, "WPers")
    """Ramp rate in watts per second."""

    LPERS = (68, "lPers")
    """Volumetric flow rate in litres per second."""

    DBM = (69, "dBm")
    """Power level (logarithmic ratio of signal strength , Bel-mW), normalized to 1mW. 
    Note: multiplier “d” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    HOURS = (70, "h")
    """Time in hours, hour = 60 min = 3600 s."""

    MIN = (71, "min")
    """Time in minutes, minute = 60 s."""

    Q = (72, "Q")
    """Quantity power, Q."""

    QH = (73, "Qh")
    """Quantity energy, Qh."""

    OHMM = (74, "ohmm")
    """Resistivity, ohm metres, (rho)."""

    APERM = (75, "APerm")
    """A/m, magnetic field strength, amperes per metre."""

    V2H = (76, "V2h")
    """Volt-squared hour, volt-squared-hours."""

    A2H = (77, "A2h")
    """Ampere-squared hour, ampere-squared hour."""

    AH = (78, "Ah")
    """Ampere-hours, ampere-hours."""

    COUNT = (79, "count")
    """Amount of substance, Counter value."""

    FT3 = (80, "ft3")
    """Volume, cubic feet."""

    M3PERH = (81, "m3Perh")
    """Volumetric flow rate, cubic metres per hour."""

    GAL = (82, "gal")
    """Volume in gallons, US gallon (1 gal = 231 in3 = 128 fl ounce)."""

    BTU = (83, "Btu")
    """Energy, British Thermal Units."""

    L = (84, "l")
    """Volume in litres, litre = dm3 = m3/1000."""

    LPERH = (85, "lPerh")
    """Volumetric flow rate, litres per hour."""

    LPERL = (86, "lPerl")
    """Concentration, The ratio of the volume of a solute divided by the volume of the solution. 
    Note: Users may need use a prefix such a ‘µ’ to express a quantity such as ‘µL/L’."""

    GPERG = (87, "gPerg")
    """Concentration, The ratio of the mass of a solute divided by the mass of the solution. 
    Note: Users may need use a prefix such a ‘µ’ to express a quantity such as ‘µg/g’."""

    MOLPERM3 = (88, "molPerm3")
    """Concentration, The amount of substance concentration, (c), the amount of solvent in moles divided by the volume of solution in m³."""

    MOLPERMOL = (89, "molPermol")
    """Concentration, Molar fraction, the ratio of the molar amount of a solute divided by the molar amount of the solution."""

    MOLPERKG = (90, "molPerkg")
    """Concentration, Molality, the amount of solute in moles and the amount of solvent in kilograms."""

    SPERS = (91, "sPers")
    """Time, Ratio of time. Note: Users may need to supply a prefix such as ‘µ’ to show rates such as ‘µs/s’."""

    HZPERHZ = (92, "HzPerHz")
    """Frequency, rate of frequency change. Note: Users may need to supply a prefix such as ‘m’ to show rates such as ‘mHz/Hz’."""

    VPERV = (93, "VPerV")
    """Voltage, ratio of voltages. Note: Users may need to supply a prefix such as ‘m’ to show rates such as ‘mV/V’."""

    APERA = (94, "APerA")
    """Current, ratio of amperages. Note: Users may need to supply a prefix such as ‘m’ to show rates such as ‘mA/A’."""

    VPERVA = (95, "VPerVA")
    """Power factor, PF, the ratio of the active power to the apparent power. 
    Note: The sign convention used for power factor will differ between IEC meters and EEI (ANSI) meters. 
    It is assumed that the data consumers understand the type of meter being used and agree on the sign convention in use at any given utility."""

    REV = (96, "rev")
    """Amount of rotation, revolutions."""

    KAT = (97, "kat")
    """Catalytic activity, katal = mol / s."""

    JPERKG = (98, "JPerkg")
    """Specific energy, Joules / kg."""

    M3UNCOMPENSATED = (99, "m3Uncompensated")
    """Volume, cubic metres, with the value uncompensated for weather effects."""

    M3COMPENSATED = (100, "m3Compensated")
    """Volume, cubic metres, with the value compensated for weather effects."""

    WPERW = (101, "WPerW")
    """Signal Strength, ratio of power. Note: Users may need to supply a prefix such as ‘m’ to show rates such as ‘mW/W’."""

    THERM = (102, "therm")
    """Energy, therms."""

    ONEPERM = (103, "onePerm")
    """Wavenumber, reciprocal metres, (1/m)."""

    M3PERKG = (104, "m3Perkg")
    """Specific volume, cubic metres per kilogram, v."""

    PAS = (105, "Pas")
    """Dynamic viscosity, pascal seconds."""

    NM = (106, "Nm")
    """Moment of force, newton metres."""

    NPERM = (107, "NPerm")
    """Surface tension, newton per metre."""

    RADPERS2 = (108, "radPers2")
    """Angular acceleration, radians per second squared."""

    JPERM3 = (109, "JPerm3")
    """Energy density, joules per cubic metre."""

    VPERM = (110, "VPerm")
    """Electric field strength, volts per metre."""

    CPERM3 = (111, "CPerm3")
    """Electric charge density, coulombs per cubic metre."""

    CPERM2 = (112, "CPerm2")
    """Surface charge density, coulombs per square metre."""

    FPERM = (113, "FPerm")
    """Permittivity, farads per metre."""

    HPERM = (114, "HPerm")
    """Permeability, henrys per metre."""

    JPERMOL = (115, "JPermol")
    """Molar energy, joules per mole."""

    JPERMOLK = (116, "JPermolK")
    """Molar entropy, molar heat capacity, joules per mole kelvin."""

    CPERKG = (117, "CPerkg")
    """Exposure (x rays), coulombs per kilogram."""

    GYPERS = (118, "GyPers")
    """Absorbed dose rate, grays per second."""

    WPERSR = (119, "WPersr")
    """Radiant intensity, watts per steradian."""

    WPERM2SR = (120, "WPerm2sr")
    """Radiance, watts per square metre steradian."""

    KATPERM3 = (121, "katPerm3")
    """Catalytic activity concentration, katals per cubic metre."""

    D = (122, "d")
    """Time in days, day = 24 h = 86400 s."""

    ANGLEMIN = (123, "anglemin")
    """Plane angle, minutes."""

    ANGLESEC = (124, "anglesec")
    """Plane angle, seconds."""

    HA = (125, "ha")
    """Area, hectares."""

    TONNE = (126, "tonne")
    """Mass in tons, “tonne” or “metric ton” (1000 kg = 1 Mg)."""

    BAR = (127, "bar")
    """Pressure in bars, (1 bar = 100 kPa)."""

    MMHG = (128, "mmHg")
    """Pressure, millimetres of mercury (1 mmHg is approximately 133.3 Pa)."""

    MILES_NAUTICAL = (129, "M")
    """Length, nautical miles (1 M = 1852 m)."""

    KN = (130, "kn")
    """Speed, knots (1 kn = 1852/3600) m/s."""

    MX = (131, "Mx")
    """Magnetic flux, maxwells (1 Mx = 10-8 Wb)."""

    G = (132, "G")
    """Magnetic flux density, gausses (1 G = 10-4 T)."""

    OE = (133, "Oe")
    """Magnetic field in oersteds, (1 Oe = (103/4p) A/m)."""

    VH = (134, "Vh")
    """Volt-hour, Volt hours."""

    WPERA = (135, "WPerA")
    """Active power per current flow, watts per Ampere."""

    ONEPERHZ = (136, "onePerHz")
    """Reciprocal of frequency (1/Hz)."""

    VPERVAR = (137, "VPerVAr")
    """Power factor, PF, the ratio of the active power to the apparent power. 
    Note: The sign convention used for power factor will differ between IEC meters and EEI (ANSI) meters. 
    It is assumed that the data consumers understand the type of meter being used and agree on the sign convention in use at any given utility."""

    OHMPERM = (138, "ohmPerm")
    """Electric resistance per length in ohms per metre ((V/A)/m)."""

    KGPERJ = (139, "kgPerJ")
    """Weight per energy in kilograms per joule (kg/J). Note: multiplier “k” is included in this unit symbol for compatibility with IEC 61850-7-3."""

    JPERS = (140, "JPers")
    """Energy rate in joules per second (J/s)."""

    @property
    def short_name(self):
        return str(self)[11:]

    def __str__(self):
        return self.value[1]

    def id(self):
        return self.value[0]


_unitsymbol_members_by_id = [us for us in UnitSymbol.__members__.values()]
_unitsymbol_by_cim_name = {str(us): us for us in UnitSymbol.__members__.values()}



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional
from zepben.evolve.model.cim.iec61970.base.core.identified_object import IdentifiedObject

__all__ = ["OrganisationRole"]


class OrganisationRole(IdentifiedObject):
    """
    Identifies a way in which an organisation may participate in the utility enterprise (e.g., customer, manufacturer, etc).
    """

    organisation: Optional[Organisation] = None
    """The `zepben.evolve.cim.iec61968.common.organisation.Organisation` having this role."""



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from zepben.evolve.services.network.tracing.traversals.tracker import *
from zepben.evolve.services.network.tracing.traversals.tracing import *
from zepben.evolve.services.network.tracing.traversals.queue import *
from zepben.evolve.services.network.tracing.traversals.branch_recursive_tracing import *



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional

from zepben.evolve.model.cim.iec61970.base.core.power_system_resource import PowerSystemResource
from zepben.evolve.model.cim.iec61970.base.wires.single_phase_kind import SinglePhaseKind

__all__ = ["EnergySourcePhase"]


class EnergySourcePhase(PowerSystemResource):
    """
    A single phase of an energy source.
    """

    _energy_source: Optional[EnergySource] = None
    """The `zepben.evolve.cim.iec61970.wires.EnergySource` with this `EnergySourcePhase`"""

    phase: SinglePhaseKind = SinglePhaseKind.NONE
    """A `zepben.evolve.iec61970.base.wires.single_phase_kind.SinglePhaseKind` Phase of this energy source component. If the energy source is wye connected, 
    the connection is from the indicated phase to the central ground or neutral point. If the energy source is delta connected, the phase indicates an energy 
    source connected from the indicated phase to the next logical non-neutral phase."""

    def __init__(self, energy_source: EnergySource = None):
        if energy_source:
            self.energy_source = energy_source

    @property
    def energy_source(self):
        """The `zepben.evolve.cim.iec61970.wires.energy_source.EnergySource` with this `EnergySourcePhase`"""
        return self._energy_source

    @energy_source.setter
    def energy_source(self, es):
        if self._energy_source is None or self._energy_source is es:
            self._energy_source = es
        else:
            raise ValueError(f"energy_source for {str(self)} has already been set to {self._energy_source}, cannot reset this field to {es}")



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from zepben.protobuf.cim.iec61970.base.meas.MeasurementValue_pb2 import MeasurementValue as PBMeasurementValue
from zepben.protobuf.cim.iec61970.base.meas.AccumulatorValue_pb2 import AccumulatorValue as PBAccumulatorValue
from zepben.protobuf.cim.iec61970.base.meas.AnalogValue_pb2 import AnalogValue as PBAnalogValue
from zepben.protobuf.cim.iec61970.base.meas.DiscreteValue_pb2 import DiscreteValue as PBDiscreteValue
from zepben.evolve.model import AnalogValue, AccumulatorValue, DiscreteValue, MeasurementValue
from google.protobuf.timestamp_pb2 import Timestamp

__all__ = ["analogvalue_to_pb", "accumulatorvalue_to_pb", "discretevalue_to_pb", "measurementvalue_to_pb"]


def analogvalue_to_pb(cim: AnalogValue) -> PBAnalogValue:
    return PBAnalogValue(mv=measurementvalue_to_pb(cim), analogMRID=cim.analog_mrid, value=cim.value)


def accumulatorvalue_to_pb(cim: AccumulatorValue) -> PBAccumulatorValue:
    return PBAccumulatorValue(mv=measurementvalue_to_pb(cim), accumulatorMRID=cim.accumulator_mrid, value=cim.value)


def discretevalue_to_pb(cim: DiscreteValue) -> PBDiscreteValue:
    return PBDiscreteValue(mv=measurementvalue_to_pb(cim), discreteMRID=cim.discrete_mrid, value=cim.value)


def measurementvalue_to_pb(cim: MeasurementValue) -> PBMeasurementValue:
    ts = Timestamp()
    ts.FromDatetime(cim.time_stamp)
    return PBMeasurementValue(timeStamp=ts)


AnalogValue.to_pb = analogvalue_to_pb
AccumulatorValue.to_pb = accumulatorvalue_to_pb
DiscreteValue.to_pb = discretevalue_to_pb
MeasurementValue.to_pb = measurementvalue_to_pb




#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Optional, Generator, List

from zepben.evolve.model.cim.iec61970.base.core.equipment_container import Feeder, Site
from zepben.evolve.model.cim.iec61970.base.core.power_system_resource import PowerSystemResource
from zepben.evolve.model.cim.iec61970.base.core.substation import Substation
from zepben.evolve.util import nlen, get_by_mrid, ngen, safe_remove

__all__ = ['Equipment']


class Equipment(PowerSystemResource):
    """
    Abstract class, should only be used through subclasses.
    Any part of a power system that is a physical device, electronic or mechanical.
    """

    in_service: bool = True
    """If True, the equipment is in service."""
    normally_in_service: bool = True
    """If True, the equipment is _normally_ in service."""

    _usage_points: Optional[List[UsagePoint]] = None
    _equipment_containers: Optional[List[EquipmentContainer]] = None
    _operational_restrictions: Optional[List[OperationalRestriction]] = None
    _current_feeders: Optional[List[Feeder]] = None

    def __init__(self, usage_points: List[UsagePoint] = None, equipment_containers: List[EquipmentContainer] = None,
                 operational_restrictions: List[OperationalRestriction] = None, current_feeders: List[Feeder] = None):
        if usage_points:
            for up in usage_points:
                self.add_usage_point(up)
        if equipment_containers:
            for container in equipment_containers:
                self.add_container(container)
        if operational_restrictions:
            for restriction in operational_restrictions:
                self.add_restriction(restriction)
        if current_feeders:
            for cf in current_feeders:
                self.add_current_feeder(cf)

    @property
    def equipment_containers(self) -> Generator[Equipment, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer`s this equipment belongs to.
        """
        return ngen(self._equipment_containers)

    @property
    def current_feeders(self) -> Generator[Feeder, None, None]:
        """
        The current `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`s this equipment belongs to.
        """
        return ngen(self._current_feeders)

    @property
    def normal_feeders(self) -> Generator[Feeder, None, None]:
        """
        The normal `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`s this equipment belongs to.
        """
        for feeder in self._equipment_containers_of_type(Feeder):
            yield feeder

    @property
    def sites(self) -> Generator[Site, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.equipment_container.Site`s this equipment belongs to.
        """
        for site in self._equipment_containers_of_type(Site):
            yield site

    @property
    def substations(self) -> Generator[Substation, None, None]:
        """
        The `zepben.evolve.cim.iec61970.base.core.substation.Substation`s this equipment belongs to.
        """
        for sub in self._equipment_containers_of_type(Substation):
            yield sub

    @property
    def usage_points(self) -> Generator[UsagePoint, None, None]:
        """
        The `zepben.evolve.cim.iec61968.metering.metering.UsagePoint`s for this equipment.
        """
        return ngen(self._usage_points)

    @property
    def operational_restrictions(self) -> Generator[OperationalRestriction, None, None]:
        """
        The `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction`s that this equipment is associated with.
        """
        return ngen(self._operational_restrictions)

    def num_equipment_containers(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer`s associated with this `Equipment`
        """
        return nlen(self._equipment_containers)

    def num_substations(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61970.base.core.substation.Substation`s associated with this `Equipment`
        """
        return len(self._equipment_containers_of_type(Substation))

    def num_sites(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61970.base.core.equipment_container.Site`s associated with this `Equipment`
        """
        return len(self._equipment_containers_of_type(Site))

    def num_normal_feeders(self) -> int:
        """
        Returns The number of normal `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`s associated with this `Equipment`
        """
        return len(self._equipment_containers_of_type(Feeder))

    def num_usage_points(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61968.metering.metering.UsagePoint`s associated with this `Equipment`
        """
        return nlen(self._usage_points)

    def num_current_feeders(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`s associated with this `Equipment`
        """
        return nlen(self._current_feeders)

    def num_restrictions(self) -> int:
        """
        Returns The number of `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction`s associated with this `Equipment`
        """
        return nlen(self._operational_restrictions)

    def get_container(self, mrid: str) -> Equipment:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer` for this `Equipment` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer`
        Returns The `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._equipment_containers, mrid)

    def add_container(self, ec: EquipmentContainer) -> Equipment:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer` with this `Equipment`

        `ec` The `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer` to associate with this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if another `EquipmentContainer` with the same `mrid` already exists for this `Equipment`.
        """
        if self._validate_reference(ec, self.get_container, "An EquipmentContainer"):
            return self
        self._equipment_containers = list() if self._equipment_containers is None else self._equipment_containers
        self._equipment_containers.append(ec)
        return self

    def remove_containers(self, ec: EquipmentContainer) -> Equipment:
        """
        Disassociate `ec` from this `Equipment`.

        `ec` The `zepben.evolve.cim.iec61970.base.core.equipment_container.EquipmentContainer` to disassociate from this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if `ec` was not associated with this `Equipment`.
        """
        self._equipment_containers = safe_remove(self._equipment_containers, ec)
        return self

    def clear_containers(self) -> Equipment:
        """
        Clear all equipment.
        Returns A reference to this `Equipment` to allow fluent use.
        """
        self._equipment_containers = None
        return self

    def get_current_feeder(self, mrid: str) -> Equipment:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` for this `Equipment` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder`
        Returns The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._current_feeders, mrid)

    def add_current_feeder(self, feeder: Feeder) -> Equipment:
        """
        Associate `feeder` with this `Equipment`.

        `feeder` The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` to associate with this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if another `Feeder` with the same `mrid` already exists for this `Equipment`.
        """
        if self._validate_reference(feeder, self.get_current_feeder, "A Feeder"):
            return self
        self._current_feeders = list() if self._current_feeders is None else self._current_feeders
        self._current_feeders.append(feeder)
        return self

    def remove_current_feeder(self, feeder: Feeder) -> Equipment:
        """
        Disassociate `feeder` from this `Equipment`

        `feeder` The `zepben.evolve.cim.iec61970.base.core.equipment_container.Feeder` to disassociate from this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if `feeder` was not associated with this `Equipment`.
        """
        self._current_feeders = safe_remove(self._current_feeders, feeder)
        return self

    def clear_current_feeders(self) -> Equipment:
        """
        Clear all current `Feeder`s.
        Returns A reference to this `Equipment` to allow fluent use.
        """
        self._current_feeders = None
        return self

    def get_usage_point(self, mrid: str) -> UsagePoint:
        """
        Get the `zepben.evolve.cim.iec61968.metering.metering.UsagePoint` for this `Equipment` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61968.metering.metering.UsagePoint`
        Returns The `zepben.evolve.cim.iec61968.metering.metering.UsagePoint` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._usage_points, mrid)

    def add_usage_point(self, up: UsagePoint) -> Equipment:
        """
        Associate `up` with this `Equipment`.

        `up` the `zepben.evolve.cim.iec61968.metering.metering.UsagePoint` to associate with this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if another `UsagePoint` with the same `mrid` already exists for this `Equipment`.
        """
        if self._validate_reference(up, self.get_usage_point, "A UsagePoint"):
            return self
        self._usage_points = list() if self._usage_points is None else self._usage_points
        self._usage_points.append(up)
        return self

    def remove_usage_point(self, up: UsagePoint) -> Equipment:
        """
        Disassociate `up` from this `Equipment`.

        `up` The `zepben.evolve.cim.iec61968.metering.metering.UsagePoint` to disassociate from this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if `up` was not associated with this `Equipment`.
        """
        self._usage_points = safe_remove(self._usage_points, up)
        return self

    def clear_usage_points(self) -> Equipment:
        """
        Clear all usage_points.
        Returns A reference to this `Equipment` to allow fluent use.
        """
        self._usage_points = None
        return self

    def get_restriction(self, mrid: str) -> OperationalRestriction:
        """
        Get the `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction` for this `Equipment` identified by `mrid`

        `mrid` The mRID of the required `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction`
        Returns The `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._operational_restrictions, mrid)

    def add_restriction(self, op: OperationalRestriction) -> Equipment:
        """
        Associate `op` with this `Equipment`.

        `op` The `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction` to associate with this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if another `OperationalRestriction` with the same `mrid` already exists for this `Equipment`.
        """
        if self._validate_reference(op, self.get_restriction, "An OperationalRestriction"):
            return self
        self._operational_restrictions = list() if self._operational_restrictions is None else self._operational_restrictions
        self._operational_restrictions.append(op)
        return self

    def remove_restriction(self, op: OperationalRestriction) -> Equipment:
        """
        Disassociate `up` from this `Equipment`.

        `op` The `zepben.evolve.cim.iec61968.operations.operational_restriction.OperationalRestriction` to disassociate from this `Equipment`.
        Returns A reference to this `Equipment` to allow fluent use.
        Raises `ValueError` if `op` was not associated with this `Equipment`.
        """
        self._operational_restrictions = safe_remove(self._operational_restrictions, op)
        return self

    def clear_restrictions(self) -> Equipment:
        """
        Clear all `OperationalRestrictions`.
        Returns A reference to this `Equipment` to allow fluent use.
        """
        self._operational_restrictions = None
        return self

    def _equipment_containers_of_type(self, ectype: type) -> List[EquipmentContainer]:
        """Get the `EquipmentContainer`s for this `Equipment` of type `ectype`"""
        if self._equipment_containers:
            return [ec for ec in self._equipment_containers if isinstance(ec, ectype)]
        else:
            return []



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations
from typing import Optional, Generator, List

from zepben.evolve.model import Line
from zepben.evolve.util import ngen, get_by_mrid, safe_remove, nlen

__all__ = ["Circuit"]


class Circuit(Line):
    """Missing description"""

    loop: Optional[Loop] = None
    _end_terminals: Optional[List[Terminal]] = None
    _end_substations: Optional[List[Substation]] = None

    def __init__(self, equipment: List[Equipment] = None, end_terminals: List[Terminal] = None, end_substations: List[Substation] = None):
        super(Circuit, self).__init__(equipment)
        if end_terminals:
            for term in end_terminals:
                self.add_end_terminal(term)

        if end_substations:
            for sub in end_substations:
                self.add_end_substation(sub)

    @property
    def end_terminals(self) -> Generator[Terminal, None, None]:
        """
        The `Terminal`s representing the ends for this `Circuit`.
        """
        return ngen(self._end_terminals)

    @property
    def end_substations(self) -> Generator[Substation, None, None]:
        """
        The `Substations`s representing the ends for this `Circuit`.
        """
        return ngen(self._end_substations)

    def num_end_terminals(self):
        """Return the number of end `Terminal`s associated with this `Circuit`"""
        return nlen(self._end_terminals)

    def get_terminal(self, mrid: str) -> Circuit:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` for this `Circuit` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.core.terminal.Terminal`
        Returns The `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._end_terminals, mrid)

    def add_terminal(self, terminal: Terminal) -> Circuit:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` with this `Circuit`

        `terminal` the `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to associate with this `Circuit`.
        Returns A reference to this `Circuit` to allow fluent use.
        Raises `ValueError` if another `Terminal` with the same `mrid` already exists for this `Circuit`.
        """
        if self._validate_reference(terminal, self.get_terminal, "An Terminal"):
            return self
        self._end_terminals = list() if self._end_terminals is None else self._end_terminals
        self._end_terminals.append(terminal)
        return self

    def remove_end_terminals(self, terminal: Terminal) -> Circuit:
        """
        Disassociate `terminal` from this `Circuit`

        `terminal` the `zepben.evolve.cim.iec61970.base.core.terminal.Terminal` to disassociate from this `Circuit`.
        Returns A reference to this `Circuit` to allow fluent use.
        Raises `ValueError` if `terminal` was not associated with this `Circuit`.
        """
        self._end_terminals = safe_remove(self._end_terminals, terminal)
        return self

    def clear_end_terminals(self) -> Circuit:
        """
        Clear all end terminals.
        Returns A reference to this `Circuit` to allow fluent use.
        """
        self._end_terminals = None
        return self

    def num_end_substations(self):
        """Return the number of end `Substation`s associated with this `Circuit`"""
        return nlen(self._end_substations)

    def get_substation(self, mrid: str) -> Circuit:
        """
        Get the `zepben.evolve.cim.iec61970.base.core.substation.Substation` for this `Circuit` identified by `mrid`

        `mrid` the mRID of the required `zepben.evolve.cim.iec61970.base.core.substation.Substation`
        Returns The `zepben.evolve.cim.iec61970.base.core.substation.Substation` with the specified `mrid` if it exists
        Raises `KeyError` if `mrid` wasn't present.
        """
        return get_by_mrid(self._end_substations, mrid)

    def add_substation(self, substation: Substation) -> Circuit:
        """
        Associate an `zepben.evolve.cim.iec61970.base.core.substation.Substation` with this `Circuit`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to associate with this `Circuit`.
        Returns A reference to this `Circuit` to allow fluent use.
        Raises `ValueError` if another `Substation` with the same `mrid` already exists for this `Circuit`.
        """
        if self._validate_reference(substation, self.get_substation, "An Substation"):
            return self
        self._end_substations = list() if self._end_substations is None else self._end_substations
        self._end_substations.append(substation)
        return self

    def remove_end_substations(self, substation: Substation) -> Circuit:
        """
        Disassociate `substation` from this `Circuit`

        `substation` the `zepben.evolve.cim.iec61970.base.core.substation.Substation` to disassociate from this `Circuit`.
        Returns A reference to this `Circuit` to allow fluent use.
        Raises `ValueError` if `substation` was not associated with this `Circuit`.
        """
        self._end_substations = safe_remove(self._end_substations, substation)
        return self

    def clear_end_substations(self) -> Circuit:
        """
        Clear all end substations.
        Returns A reference to this `Circuit` to allow fluent use.
        """
        self._end_substations = None
        return self



#  Copyright 2020 Zeppelin Bend Pty Ltd
#
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum

__all__ = ["PhaseDirection", "phasedirection_from_value"]


class PhaseDirection(Enum):
    NONE = 0
    IN = 1
    OUT = 2
    BOTH = 3

    def has(self, other):
        """
        Check whether this Direction contains Direction other.
        `other` A `Direction` to compare against.
        Returns True if this is BOTH and other is not NONE, otherwise False
        """
        if self is PhaseDirection.BOTH:
            return other is not PhaseDirection.NONE
        else:
            return self is other

    def __add__(self, other):
        return PhaseDirection(self.value | other.value)

    def __sub__(self, other):
        return PhaseDirection(self.value - (self.value and other.value))

    @property
    def short_name(self):
        return str(self)[10:]


def phasedirection_from_value(value: int) -> PhaseDirection:
    if 0 <= value <= 3:
        return PhaseDirection.NONE
    else:
        PhaseDirection(value)
